import pandas as pd
import plotnine as p9

from liana._constants import Keys as K, DefaultValues as V
from liana._docs import d

@d.dedent
def target_metrics(misty = None,
                   stat = None,
                   target_metrics = None,
                   top_n = None,
                   ascending = False,
                   key = None,
                   filter_fun: callable = None,
                   figure_size: tuple = (5, 5),
                   aggregate_fun = None,
                   return_fig: bool = V.return_fig
                   ):
    """
    Plot target metrics.

    Parameters
    ----------
    %(misty)s
    stat : str
        Statistic to plot
    target_metrics
        A target_metrics DataFrame
    %(top_n)s
    ascending : bool
        Whether to sort in ascending order
    key : callable
        Function to use to sort the dataframe
    %(filter_fun)s
    %(aggregate_fun)s
    %(figure_size)s
    %(return_fig)s

    Returns
    -------

    Returns a plotnine plot.

    """
    if target_metrics is not None:
        target_metrics = target_metrics.copy()
    elif misty is not None:
        target_metrics = misty.uns[K.target_metrics].copy()
    else:
        raise ValueError("Provide either a misty object or a target_metrics DataFrame.")
    if stat is None:
        raise ValueError("Provide a statistic to plot")

    if filter_fun is not None:
        target_metrics = target_metrics[target_metrics.apply(filter_fun, axis=1).astype(bool)]
    if aggregate_fun is not None:
        targets = target_metrics.groupby(['target']).agg({stat: aggregate_fun})
        targets = targets.sort_values(by=stat, ascending=ascending).index
    else:
        targets = target_metrics.sort_values(by=stat, ascending=ascending, key=key)['target'].unique()
    if top_n is not None:
        target_metrics = target_metrics[target_metrics['target'].isin(targets[:top_n])]

    # keep order of targets
    target_metrics['target'] = pd.Categorical(target_metrics['target'],
                                              categories=targets,
                                              ordered=True)

    p = (p9.ggplot(target_metrics, p9.aes(x='target', y=stat)) +
         (p9.geom_boxplot() if aggregate_fun is not None else p9.geom_point(size=3)) +
         p9.theme_bw() +
         p9.theme(axis_text_x=p9.element_text(rotation=90),
                  figure_size=figure_size) +
         p9.labs(x='Target')
         )

    if return_fig:
        return p
    p.draw()

@d.dedent
def contributions(misty = None,
                  target_metrics = None,
                  view_names: list = None,
                  filter_fun: callable = None,
                  aggregate_fun: callable = None,
                  figure_size: tuple = (5, 5),
                  return_fig: bool = V.return_fig):
    """
    Plot view contributions per target.

    Parameters
    ----------

    %(misty)s
    target_metrics
        A target_metrics DataFrame
    view_names
        A list of view names to plot
    %(aggregate_fun)s
    %(figure_size)s
    %(return_fig)s

    Returns
    -------
    A plotnine plot.

    """
    if target_metrics is not None:
        target_metrics = target_metrics.copy()
    elif misty is not None:
        target_metrics = misty.uns[K.target_metrics].copy()
    else:
        raise ValueError("Provide either a misty object or a target_metrics DataFrame.")

    if view_names is None:
        if misty is None:
            raise ValueError("Provide a list of view names to plot.")
        view_names = misty.view_names.copy()
        if 'intra' not in target_metrics.columns:
            view_names.remove('intra')

    if filter_fun is not None:
        target_metrics = target_metrics[target_metrics.apply(filter_fun, axis=1).astype(bool)]
        target_metrics['target'] = target_metrics['target'].cat.remove_unused_categories()

    target_metrics = target_metrics[['target', *view_names]]
    target_metrics = target_metrics.melt(id_vars='target', var_name='view', value_name='contribution')

    if aggregate_fun is not None:
        target_metrics = target_metrics.groupby(['target', 'view']).agg({'contribution': aggregate_fun}).reset_index()

    p = (p9.ggplot(target_metrics, p9.aes(x='target', y='contribution', fill='view')) +
            p9.geom_bar(stat='identity') +
            p9.theme_bw(base_size=14) +
            p9.theme(axis_text_x=p9.element_text(rotation=90),
                     figure_size=figure_size) +
            p9.scale_fill_brewer(palette=2, type='qual') +
            p9.labs(x='Target', y='Contribution', fill='View')
    )

    if return_fig:
        return p
    p.draw()


@d.dedent
def interactions(misty= None,
                 interactions = None,
                 view = None,
                 top_n = None,
                 ascending = False,
                 key = None,
                 filter_fun: callable = None,
                 aggregate_fun: callable = None,
                 figure_size: tuple = (5,5),
                 return_fig: bool = V.return_fig):
    """
    Plot interaction importances.

    Parameters
    ----------

    %(misty)s
    view : str
        A view to plot
    %(top_n)s
    ascending : bool
        Whether to sort interactions in ascending order
    key : str
        Key to use when sorting interactions
    %(filter_fun)s
    %(aggregate_fun)s
    %(figure_size)s
    %(return_fig)s

    Returns
    -------
    A plotnine plot.

    """
    if interactions is not None:
        interactions = interactions.copy()
    elif misty is not None:
        interactions = misty.uns[K.interactions].copy()
    else:
        raise ValueError("Provide either a misty object or interactions.")
    if view is None:
        raise ValueError("Provide a ``view`` to plot!")

    interactions = interactions[interactions['view'] == view]
    grouped = interactions.groupby('predictor')['importances'].apply(lambda x: x.isna().all())
    interactions = interactions[~interactions['predictor'].isin(grouped[grouped].index)]

    if filter_fun is not None:
        interactions = interactions[interactions.apply(filter_fun, axis=1).astype(bool)]
    if aggregate_fun is not None:
        interactions = interactions.groupby(['target', 'predictor']).agg({'importances': aggregate_fun}).reset_index()
    if top_n is not None:
        interactions = interactions.sort_values(by='importances', key=key, ascending=ascending)
        top_interactions = interactions.drop_duplicates(['target', 'predictor']).head(top_n)
        interactions = interactions[interactions['target'].isin(top_interactions['target']) &
                            interactions['predictor'].isin(top_interactions['predictor'])]

    p = (p9.ggplot(interactions,
                   p9.aes(x='predictor',
                          y='target',
                          fill='importances')
                   ) +
    p9.geom_tile() +
    p9.theme_minimal(base_size=12) +
    p9.theme(axis_text_x=p9.element_text(rotation=90),
             figure_size=figure_size) +
    p9.labs(x='Predictor', y='Target', fill='Importance')
    )

    if return_fig:
        return p
    p.draw()



import numpy as np
import pandas as pd

from liana._constants import Keys as K

def _check_var(liana_res, var_name, var):
    if var is None:
        raise ValueError(f'`{var_name}` must be provided!')
    if var not in liana_res.columns:
        raise ValueError(f'`{var}` ({var_name}) must be one of {liana_res.columns}')


def _prep_liana_res(adata=None,
                    liana_res=None,
                    source_labels=None,
                    target_labels=None,
                    ligand_complex=None,
                    receptor_complex=None,
                    uns_key=K.uns_key):

    if (liana_res is None) & (adata is None):
        raise AttributeError(f'Ambiguous! One of `liana_res` or `adata.uns[{uns_key}]` should be provided.')
    if adata is not None:
        assert uns_key in adata.uns.keys()
        liana_res = adata.uns[uns_key].copy()
    if liana_res is not None:
        liana_res = liana_res.copy()
    if (liana_res is None) & (adata is None):
        raise ValueError('`liana_res` or `adata` must be provided!')

    # subset to only cell labels of interest
    liana_res = _filter_labels(liana_res, labels=source_labels, label_type='source')
    liana_res = _filter_labels(liana_res, labels=target_labels, label_type='target')

    liana_res['interaction'] = liana_res['ligand_complex'] + ' -> ' + liana_res['receptor_complex']

    if ligand_complex is not None:
        liana_res = liana_res[np.isin(liana_res['ligand_complex'], ligand_complex)]
    if receptor_complex is not None:
        liana_res = liana_res[np.isin(liana_res['receptor_complex'], receptor_complex)]

    return liana_res


def _filter_labels(liana_res, labels, label_type):
    if labels is not None:
        if labels is str:
            labels = [labels]
        covered = np.isin(labels, liana_res[label_type])
        if not covered.all():
            not_covered = np.array(labels)[~covered]
            raise ValueError(f"{not_covered} not found in `liana_res['{label_type}']`!")
        msk = np.isin(liana_res[label_type], labels)
        liana_res = liana_res[msk]

    return liana_res


def _aggregate_scores(res, what, how, absolute, entities):
    res['score'] = np.absolute(res[what]) if absolute else res[what]
    res = res.groupby(entities).agg(score=('score', how)).reset_index()
    return res


def _inverse_scores(score):
    return -np.log10(score + np.finfo(float).eps)


def _filter_by(liana_res, filter_fun):
    if filter_fun is not None:
        msk = liana_res.apply(filter_fun, axis=1).astype(bool)
        relevant_interactions = np.unique(liana_res[msk].interaction)
        liana_res = liana_res[np.isin(liana_res['interaction'], relevant_interactions)]

    return liana_res


def _get_top_n(liana_res, top_n, orderby, orderby_ascending, orderby_absolute):

    if top_n is not None:
        # get the top_n for each interaction
        if orderby is None:
            ValueError("Please specify the column to order the interactions.")
        if orderby_ascending is None:
            ValueError("Please specify if `orderby` is ascending or not.")
        if orderby_ascending:
            how = 'min'
        else:
            how = 'max'

        top_lrs = _aggregate_scores(liana_res,
                                    what=orderby,
                                    how=how,
                                    absolute=orderby_absolute,
                                    entities=['interaction',
                                              'ligand_complex',
                                              'receptor_complex']
                                    ).copy()
        top_lrs = top_lrs.sort_values('score', ascending=orderby_ascending).head(top_n).interaction

        # Filter liana_res to the interactions in top_lrs
        liana_res = liana_res[liana_res['interaction'].isin(top_lrs)]
        # set categories to the order of top_lrs
        liana_res['interaction'] = pd.Categorical(liana_res['interaction'], categories=top_lrs)

    return liana_res



import plotnine as p9
from typing import Union, List, Tuple
import anndata as ad
import pandas as pd

from liana.plotting._common import _prep_liana_res, _get_top_n, _check_var, _filter_by
from liana._docs import d
from liana._constants import Keys as K, DefaultValues as V

@d.dedent
def tileplot(adata: ad.AnnData = None,
             liana_res: pd.DataFrame = None,
             fill: str = None,
             label: str = None,
             label_fun: callable = None,
             source_labels: Union[str, List[str]] = None,
             target_labels: Union[str, List[str]] = None,
             ligand_complex: Union[str, List[str]] = None,
             receptor_complex: Union[str, List[str]] = None,
             uns_key: str = K.uns_key,
             top_n: int = None,
             orderby: str = None,
             orderby_ascending: bool = False,
             orderby_absolute: bool = True,
             filter_fun: callable = None,
             source_title=None,
             target_title=None,
             cmap: str = V.cmap,
             figure_size: Tuple[float, float] = (5, 5),
             label_size: int = 12,
             return_fig: bool = V.return_fig
             ):
    """
    Tileplot interactions by source and target cells

    Parameters
    ----------
    %(adata)s
    %(liana_res)s
    fill
        `column` in `liana_res` to define the fill of the tiles
    label
        `column` in `liana_res` to define the label of the tiles
    label_fun
        `callable` to apply to the `label` column
    %(source_labels)s
    %(target_labels)s
    %(ligand_complex)s
    %(receptor_complex)s
    %(uns_key)s
    %(top_n)s
    %(orderby)s
    %(orderby_ascending)s
    %(orderby_absolute)s
    %(filter_fun)s
    source_title
        Title for the source facet. Default is 'Source'
    target_title
        Title for the target facet. Default is 'Target'
    %(cmap)s
    label_size
        Size of the label text
    %(figure_size)s
    %(return_fig)s

    Returns
    -------
    A `plotnine.ggplot` instance

    """
    liana_res = _prep_liana_res(adata=adata,
                                liana_res=liana_res,
                                source_labels=source_labels,
                                target_labels=target_labels,
                                ligand_complex=ligand_complex,
                                receptor_complex=receptor_complex,
                                uns_key=uns_key)

    liana_res = _filter_by(liana_res, filter_fun)
    liana_res = _get_top_n(liana_res, top_n, orderby, orderby_ascending, orderby_absolute)

    # get columns which ends with fill or label
    relevant_cols = [col for col in liana_res.columns if col.endswith(fill) | col.endswith(label)]

    ligand_stats = _entity_stats(liana_res,
                                 entity='ligand',
                                 entity_type='source',
                                 relevant_cols=relevant_cols,
                                 type_title=source_title)

    _check_var(ligand_stats, var=fill, var_name='fill')
    _check_var(ligand_stats, var=label, var_name='label')

    receptor_stats = _entity_stats(liana_res,
                                   entity='receptor',
                                   entity_type='target',
                                   relevant_cols=relevant_cols,
                                   type_title=target_title)

    liana_res = pd.concat([ligand_stats, receptor_stats])

    if label_fun is not None:
        liana_res[label] = liana_res[label].apply(label_fun)

    p = (
        p9.ggplot(liana_res, p9.aes(x='cell_type', y='interaction', fill=fill)) +
        p9.geom_tile() +
        p9.geom_text(p9.aes(label=label), size=label_size, color='white') +
        p9.facet_grid('~ type', scales='free') +
        p9.theme_bw(base_size=14) +
        p9.theme(
            axis_text_x=p9.element_text(angle=90),
            figure_size=figure_size,
            strip_background=p9.element_rect(colour="black", fill="#fdfff4"),
        ) +
        p9.scale_fill_cmap(cmap) +
        p9.labs(x='Cell type', y='Interaction', fill=str.capitalize(fill))
    )

    if return_fig:
        return p

    p.draw()

def _entity_stats(liana_res, entity, entity_type, relevant_cols, type_title=None):
    entity_stats = liana_res[['interaction', f"{entity}_complex", entity_type, *relevant_cols]].copy()
    if type_title is None:
        type_title = entity_type.capitalize()
    entity_stats = entity_stats.rename(columns={entity_type: 'cell_type'}).assign(type=type_title)
    entity_stats.columns = entity_stats.columns.str.replace(entity + '_', '')
    return entity_stats



from __future__ import annotations

import numpy as np
import pandas as pd
import warnings as warnings

from anndata import AnnData
from mudata import MuData
from tqdm import tqdm

from ..method import process_scores
from liana._logging import _check_if_installed
from liana.method._pipe_utils import _check_groupby
from liana._docs import d
from liana._constants import DefaultValues as V, Keys as K, PrimaryColumns as P

@d.dedent
def adata_to_views(adata: AnnData,
                   groupby: str,
                   sample_key: str,
                   obs_keys: list = None,
                   view_sep: str = ':',
                   min_count: int = 10,
                   min_total_count: int = 15,
                   large_n: int = 10,
                   min_prop: float = 0.1,
                   keep_stats: bool = False,
                   verbose: bool = False,
                   **kwargs):
    """
    Converts an AnnData object to a MuData object with views that represent an aggregate for each entity in `adata.obs[groupby]`.

    Parameters
    ----------
    %(adata)s
    %(groupby)s
    %(sample_key)s
    obs_keys:
        Column names in `adata.obs` to merge with the MuData object
    view_sep:
        Separator to use when assigning `adata.var_names` to views
    min_count:
        Minimum number of counts per gene per sample to be included in the pseudobulk.
    min_total_count:
        Minimum number of counts per sample to be included in the pseudobulk.
    large_n:
        Number of samples per group that is considered to be "large".
    min_prop:
        Minimum proportion of samples that must have a count for a gene to be included in the pseudobulk.
    keep_stats:
        If True, keep the pseudobulk statistics in `mdata.uns['psbulk_stats']`. Default is False.
    %(verbose)s
    **kwargs
        Keyword arguments used to aggregate the values per cell into views. See `dc.filter_by_expr` for more details.

    Returns
    -------
    Returns a MuData object with views that represent an aggregate for each entity in `adata.obs[groupby]`.

    """

    # Check if MuData & decoupler are installed
    dc = _check_if_installed(package_name="decoupler")

    views = adata.obs[groupby].unique()
    views = tqdm(views, disable=not verbose)

    _check_groupby(adata=adata, groupby=groupby, verbose=verbose)

    padatas = {}
    if keep_stats:
        stats = []
    for view in (views):
        # filter AnnData to view
        temp = adata[adata.obs[groupby] == view].copy()
        # assign view to var_names
        temp.var_names = view + view_sep + temp.var_names

        padata = dc.get_pseudobulk(temp,
                                   sample_col=sample_key,
                                   groups_col=None,
                                   **kwargs
                                   )

        # only filter genes for views that pass QC
        if 0 in padata.shape:
            continue

        # edgeR filtering
        feature_mask = dc.filter_by_expr(padata,
                                         min_count=min_count,
                                         min_total_count=min_total_count,
                                         large_n=large_n,
                                         min_prop=min_prop,
                                         )
        padata = padata[:, feature_mask]

        # only append views that pass QC
        if 0 not in padata.shape:
            # keep psbulk stats
            if keep_stats:
                df = padata.obs.filter(items=['psbulk_n_cells', 'psbulk_counts'], axis=1)
                df.columns = [view + view_sep + col for col in df.columns]
                stats.append(df)

            del padata.obs
            padatas[view] = padata

    # Convert to MuData
    mdata = MuData(padatas)

    # process metadata
    _process_meta(adata=adata, mdata=mdata, sample_key=sample_key, obs_keys=obs_keys)

    # combine psbulk stats across views and add to mdata
    if keep_stats:
        mdata.uns['psbulk_stats'] = pd.concat(stats, axis=1)

    return mdata

@d.dedent
def lrs_to_views(adata: AnnData,
                 score_key: (str or None) = None,
                 inverse_fun: callable = V.inverse_fun,
                 obs_keys: (list or None) = None,
                 lr_prop: float = 0.5,
                 lr_fill: np.nan = np.nan,
                 lrs_per_view:int = 20,
                 lrs_per_sample:int = 10,
                 samples_per_view: int = 3,
                 min_variance:int = 0,
                 lr_sep: str = V.lr_sep,
                 cell_sep: str='&',
                 var_sep: str=':',
                 uns_key: str = K.uns_key,
                 sample_key: str = 'sample',
                 source_key: str = P.source,
                 target_key: str = P.target,
                 ligand_key: str = P.ligand_complex,
                 receptor_key: str = P.receptor_complex,
                 verbose: bool = V.verbose
                 ):
    """
    Converts a LIANA result to a MuData object with views that represent an aggregate for each entity in `adata.obs[groupby]`.

    Parameters
    ----------
    %(adata)s
    %(score_key)s
    %(inverse_fun)s
    obs_keys
        List of keys in `adata.obs` that should be included in the MuData object.
        These columns should correspond to the number of samples in `adata.obs[sample_key]`.
    lr_prop
        Reflects the minimum required proportion of samples for an interaction to be considered for building the views.
    lr_fill
        Value to fill in for interactions that are not present in a view. Default is `np.nan`.
    lrs_per_sample
        Reflects the minimum required number of interactions in a sample to be considered when building a specific view.
    lrs_per_view
        Reflects the minimum required number of interactions in a view to be considered for building the views.
    samples_per_view
        Reflects the minimum required samples to keep a view.
    min_variance
        Reflects the minimum required variance across samples for each interaction in each view.
        NaNs are ignored when computing the variance.
    %(lr_sep)s
    cell_sep
        Separator to use for the cell names in the views.
    var_sep
        Separator to use for the variable names in the views.
    %(uns_key)s
    %(sample_key)s
    %(source_key)s
    %(target_key)s
    %(ligand_key)s
    %(receptor_key)s
    %(verbose)s

    Returns
    -------
    Returns a MuData object with views that represent an aggregate for each entity in `adata.obs[groupby]`.

    """
    if (sample_key not in adata.obs.columns) or (sample_key not in adata.uns[uns_key].columns):
        raise ValueError(f'`{sample_key}` not found in `adata.obs` or `adata.uns[uns_key]`!' +
                         'Please ensure that the sample key is present in both objects.')

    if uns_key not in adata.uns_keys():
        raise ValueError(f'`{uns_key}` not found in `adata.uns`! Please run `li.mt.rank_aggregate.by_sample` first.')

    liana_res = adata.uns[uns_key].copy()

    if (score_key is None) or (score_key not in liana_res.columns):
        raise ValueError(f"Score column `{score_key}` not found in `liana_res`")

    if isinstance(obs_keys, list):
        if any([key not in adata.obs.keys() for key in obs_keys]):
            raise ValueError(f'`{obs_keys}` not found in `adata.obs`!')
    elif obs_keys is not None:
        raise ValueError('`obs_keys` must be a list or `None`!')

    keys = np.array([sample_key, source_key, target_key, ligand_key, receptor_key])
    missing_keys = keys[[ key not in liana_res.columns for key in keys]]

    if any(missing_keys):
        raise ValueError(f'`{missing_keys}` not found in `adata.uns[{uns_key}]`! Please check your input.')

    # concat columns (needed for MOFA)
    liana_res['interaction'] = liana_res[ligand_key] + lr_sep + liana_res[receptor_key]
    liana_res['ct_pair'] = liana_res[source_key] + cell_sep + liana_res[target_key]
    liana_res = liana_res[[sample_key, 'ct_pair', 'interaction', score_key]]

    # get scores & invert if necessary
    liana_res = process_scores(liana_res=liana_res,
                                score_key=score_key,
                                inverse_fun=inverse_fun)

    # count samples per interaction
    count_pairs = (liana_res.
                   drop(columns=score_key).
                   groupby(['interaction', 'ct_pair']).
                   count().
                   rename(columns={sample_key: 'count'}).
                   reset_index()
                   )

    sample_n = liana_res[sample_key].nunique()

    # Keep only lrs above a certain proportion of samples
    count_pairs = count_pairs[count_pairs['count'] >= sample_n * lr_prop]
    liana_res = liana_res.merge(count_pairs.drop(columns='count') , how='inner')

    # Keep only samples above a certain number of LRs
    count_lrs = (liana_res.
                 drop(columns=score_key).
                 groupby([sample_key, 'ct_pair']).
                 count().
                 rename(columns={'interaction': 'count'}).
                 reset_index()
                 )
    count_lrs = count_lrs[count_lrs['count'] >= lrs_per_sample]
    liana_res = liana_res.merge(count_lrs.drop(columns='count') , how='inner')

    # convert to anndata views
    views = liana_res['ct_pair'].unique()
    views = tqdm(views, disable=not verbose)

    lr_adatas = {}
    for view in views:
        lrs_per_ct = liana_res[liana_res['ct_pair']==view]
        lrs_wide = lrs_per_ct.pivot(index='interaction',
                                    columns=sample_key,
                                    values=score_key)

        lrs_wide.index = view + var_sep + lrs_wide.index
        lrs_wide = lrs_wide.replace(np.nan, lr_fill)

        if lrs_wide.shape[0] >= lrs_per_view: # check if enough LRs
            temp = _dataframe_to_anndata(lrs_wide)

            # keep only variables with variance > min_variance
            temp = temp[:, np.nanvar(temp.X, axis=0) > min_variance]

            if (temp.shape[0] >= samples_per_view): # check if enough samples
                lr_adatas[view] = temp

    # to mdata
    mdata = MuData(lr_adatas)

    # process metadata
    _process_meta(adata=adata, mdata=mdata, sample_key=sample_key, obs_keys=obs_keys)

    return mdata


def _dataframe_to_anndata(df):
    obs = pd.DataFrame(index=df.columns)
    var = pd.DataFrame(index=df.index)
    X = np.array(df.values).T

    return AnnData(X=X, obs=obs, var=var, dtype=np.float32)


def _remove_mod_var(mdata, markers, view_sep, var_column):
    for current_mod in mdata.mod.keys():
        # markers in markers dict for each modality except for current_mod
        negative_markers = [marker for mod in markers.keys() if mod != current_mod for marker in markers[mod]]

        if current_mod not in list(markers.keys()):
            warnings.warn('no markers in dict for view: {0}'.format(current_mod), Warning)
        else:
            #keep negative_markers not in markers[current_mod] and add view_sep
            negative_markers = [current_mod + view_sep + marker for marker in negative_markers if marker not in markers[current_mod]]

        if var_column is None:
            # remove negative_markers from current_mod
            mdata.mod[current_mod] = mdata.mod[current_mod][:, ~mdata.mod[current_mod].var_names.isin(negative_markers)]
        else:
            # set negative_markers to False in current_mod
            mdata.mod[current_mod].var.loc[mdata.mod[current_mod].var_names.isin(negative_markers), var_column] = False

    mdata.update()

@d.dedent
def filter_view_markers(mdata: MuData,
                        markers: dict,
                        view_sep: str = ':',
                        var_column: str = 'highly_variable',
                        inplace: bool =False
                        ):
    """
    Used for removing potential cell type marker genes found in the background of other views and thought to be contamination.
    In each view, sets highly variable genes to False if they are in the markers dict for another view, but not if they are in the markers for the same view.


    Parameters
    ----------
    %(mdata)s
    markers :class:`dict`
        Dictionary with markers for each view. Keys are the views and values are lists of markers. Can contain markers for views that are not in mdata.mod.keys().
    view_sep :class:`str`, optional
        Separator between view and gene names. Defaults to ':'.
    var_column :class:`str`, optional
        Column in mdata.mod['some_view'].var that contains the highly variable genes. Defaults to 'highly_variable'.
        If set to ``None``, instead of setting the hvg genes to False, the hvg genes will be removed from the view.
    %(inplace)s
    """
    # check if markers is a dict
    if not isinstance(markers, dict):
        raise TypeError('markers is not a dict')

    # check that all keys in markers are lists
    if not all(isinstance(markers[mod], list) for mod in markers.keys()):
        raise TypeError('not all values in markers are lists')

    # check that var_column is in var for all modalities
    if var_column is not None:
        if not all(var_column in mdata.mod[mod].var.columns for mod in mdata.mod.keys()):
            raise ValueError('{0} is not in the columns of .var for all modalities'.format(var_column))

    if inplace:
        _remove_mod_var(mdata, markers, view_sep, var_column)
    else:
        cdata = mdata.copy()
        _remove_mod_var(cdata, markers, view_sep, var_column)
        return cdata


def _process_meta(adata, mdata, sample_key, obs_keys):
    if obs_keys is not None:
        metadata = adata.obs[[sample_key, *obs_keys]].drop_duplicates()

        sample_n = adata.obs[sample_key].nunique()
        if metadata.shape[0] != sample_n:
            raise ValueError('`obs_keys` must be unique per sample in `adata.obs`')

        mdata.obs.index.name = None
        mdata.obs = (mdata.obs.
                     reset_index().
                     rename(columns={"index":sample_key}).
                     merge(metadata).
                     set_index(sample_key)
                     )



from __future__ import annotations

from sklearn.decomposition import NMF
import numpy as np
import pandas as pd
import plotnine as p9
from tqdm import tqdm

from anndata import AnnData
from liana.method._pipe_utils._pre import _choose_mtx_rep
from liana._logging import _logg, _check_if_installed
from liana._docs import d

@d.dedent
def nmf(adata: AnnData=None,
        df: pd.DataFrame=None,
        n_components: (int or None)=None,
        k_range: range = range(1, 11),
        use_raw: bool=False,
        layer: (str or None)=None,
        inplace: bool=True,
        verbose: bool=False,
        **kwargs):
    """
    Fits NMF to an AnnData object.

    Parameters
    ----------
    %(adata)s
    n_components : int, None
        Number of components to use. If None, the number of components is estimated using the elbow method.
    k_range : range
        Range of components to test. Default: range(1, 10).
    %(use_raw)s
    %(layer)s
    %(inplace)s
    **kwargs : dict
        Keyword arguments to pass to ``sklearn.decomposition.NMF``.

    Returns
    -------
    If inplace is True, it will add ``NMF_W`` and ``NMF_H`` to the ``adata.obsm`` and ``adata.varm``.
    If n_components is None, it will also add ``nfm_errors`` and ``nfm_rank`` to ``adata.uns``.

    If inplace is False, it will return ``W`` and ``H``, and if n_components is None, it will also return ``errors`` and ``n_components``.
    If n_components is None and inplace, ``errors`` and ``n_components`` will be assigned to ``adata.uns``.
    If ``df`` is provided, inplace is always False.

    """
    if adata is not None:
        if isinstance(adata, AnnData):
            X = _choose_mtx_rep(adata, layer=layer, use_raw=use_raw)
        else :
            raise ValueError('Provide an AnnData object.')
    elif df is not None:
        X = df.values
    else:
        raise ValueError('Provide either an AnnData object or a DataFrame.')


    if n_components is None:
        errors, n_components = estimate_elbow(X, k_range=k_range, verbose=verbose, **kwargs)
        _plot_elbow(errors, n_components)
    else:
        errors, n_components = None, n_components

    nmf = NMF(n_components=n_components, **kwargs)
    W = nmf.fit_transform(X)
    H = nmf.components_.T

    inplace = inplace and (adata is not None)
    if inplace:
        adata.obsm['NMF_W'] = W
        adata.varm['NMF_H'] = H
        adata.uns['nmf_errors'] = errors
        adata.uns['nmf_rank'] = n_components

    return None if inplace else (W, H, errors, n_components)


def estimate_elbow(X, k_range, verbose=False, **kwargs):
    kn = _check_if_installed('kneed')
    errors = []
    for k in tqdm(k_range, disable=not verbose):
        error = _calculate_error(X, k, **kwargs)
        errors.append(error)

    kneedle = kn.KneeLocator(x=k_range,
                             y=errors,
                             direction='decreasing',
                             curve='convex',
                             interp_method='interp1d',
                             S=1
                             )
    rank = kneedle.knee

    _logg(f'Estimated rank: {rank}', verbose=verbose)

    errors = pd.DataFrame(errors,
                          index=list(k_range),
                          columns=['error']). \
                              reset_index().rename(columns={'index': 'k'})

    return errors, rank


def _calculate_error(X, n_components, **kwargs):
    nmf = NMF(n_components=n_components, **kwargs)
    W = nmf.fit_transform(X)
    H = nmf.components_

    Xhat = np.dot(W, H)
    error = np.mean(np.abs(X - Xhat))

    return error


def _plot_elbow(errors, n_components, x='k', y='error'):
    p = (
        p9.ggplot(errors, p9.aes(x=x, y=y)) +
        p9.geom_line() +
        p9.geom_point() +
        p9.theme_bw() +
        p9.scale_x_continuous(breaks=errors[x].values) +
        p9.labs(x='Component number (k)', y='Reconstruction error') +
        p9.geom_vline(xintercept=n_components, linetype='dashed', color='red')
    )
    p.draw()



from liana.method import find_causalnet, build_prior_network

input_pkn = [
    ("I1", 1, "N1"),
    ("N1", 1, "M1"),
    ("N1", 1, "M2"),
    ("I2", -1, "N2"),
    ("N2", -1, "M2"),
    ("N2", -1, "M1"),
    ]

input_scores = {"I1": 1, "I2": 1}
output_scores = {"M1": 1, "M2": 1}
node_weights = {"N1": 1, "N2": 1}

def test_build_prior_network():
    prior_graph = build_prior_network(input_pkn, input_scores, output_scores, verbose=True)
    assert prior_graph.num_vertices == 6
    assert prior_graph.num_edges == 6


def test_caulsalnet():
    prior_graph = build_prior_network(input_pkn, input_scores, output_scores, verbose=False)
    df_res, problem = find_causalnet(prior_graph,
                                     input_scores,
                                     output_scores,
                                     node_weights=node_weights,
                                     verbose=False,
                                     solver='scipy',
                                     seed=1337,
                                     max_runs=50,
                                     stable_runs=20,
                                     )

    assert problem.weights == [1.0, 0.01, 1.0, 1.0]
    assert df_res['source_pred_val'].values.sum() == 7.0
    assert df_res['target_pred_val'].values.sum() == 10
    assert (df_res[df_res['target_type']=='output']['target'].isin(['M1', 'M2'])).all()



def test_causalnet_noweights():
    prior_graph = build_prior_network(input_pkn, input_scores, output_scores, verbose=False)
    df_res, problem = find_causalnet(prior_graph,
                                     input_scores,
                                     output_scores,
                                     node_weights={"N1": 1, "N2": 0},
                                     verbose=False,
                                     solver='scipy',
                                     max_runs=50,
                                     stable_runs=20,
                                     )
    assert df_res['source_pred_val'].values.sum() == 8
    assert df_res['target_pred_val'].values.sum() == 9



from __future__ import annotations

import numpy as np
import pandas as pd
from tqdm import tqdm

from mudata import MuData
from sklearn.linear_model import LinearRegression, RidgeCV
from sklearn.model_selection import KFold

from liana.method.sp._misty._single_view_models import SingleViewModel

from liana._docs import d
from liana._constants import Keys as K, DefaultValues as V
from liana._logging import _logg


class MistyData(MuData):
    """MistyData Class used to construct multi-view objects"""
    @d.dedent
    def __init__(self,
                 data:(dict | MuData),
                 obs:(pd.DataFrame | None)=None,
                 spatial_key:str=K.spatial_key,
                 enforce_obs:bool=True,
                 **kwargs
                 ):
        """
        Construct a MistyData object from a dictionary of views (anndatas).

        Parameters
        ----------
        data : `dict`
            Dictionary of views (anndatas) or a MuData object. Note that only the .X attribute is used.
            An intra-view called "intra" is required.
        obs : `pd.DataFrame`
            DataFrame of observations. If None, the obs of the intra-view is used.
        %(spatial_key)s
        enforce_obs : `bool`, optional (default: True)
            If True, the number of observations in each extra-view must match the intra-view.
            Then the connectivities are stored in the .obsp attribute, while the weighted matrix is stored in .layers['weighted'].

            If False, the connectivities are stored in the .obsm attribute, while the weighted matrix is transposed and stored in .varm['weighted'].

        **kwargs
            Keyword arguments passed to the MuData Super class
        """
        if isinstance(data, MuData):
            data = data.mod

        super().__init__(data, **kwargs)
        self.view_names = list(self.mod.keys())
        self.spatial_key = spatial_key
        self.enforce_obs = enforce_obs
        self._check_views()
        self.obs = obs if obs is not None else self.mod['intra'].obs

    def _check_views(self):
        assert isinstance(self, MuData), "views must be a MuData object"
        assert "intra" in self.view_names, "views must contain an intra view"

        for view in self.view_names:
            if view=="intra":
                continue
            if self.enforce_obs:
                if f"{self.spatial_key}_connectivities" not in self.mod[view].obsp.keys():
                    raise ValueError(f"view {view} does not contain `{self.spatial_key}_connectivities` key in .obsp")
                if self.mod[view].shape[0] != self.mod['intra'].shape[0]:
                    raise ValueError(f"view {view} has {self.mod[view].shape[0]} observations, " + \
                                    f"but the intra-view has {self.mod['intra'].shape[0]} observations")
            else:
                if f"{self.spatial_key}_connectivities" not in self.mod[view].obsm.keys():
                    raise ValueError(f"view {view} does not contain `{self.spatial_key}_connectivities` key in .obsm")

            self._set_weighted_matrix(view)

    def _set_weighted_matrix(self, view_name):
        if self.enforce_obs:
            weights = self.mod[view_name].obsp[f"{self.spatial_key}_connectivities"]
            self.mod[view_name].layers['weighted'] = weights @ self.mod[view_name].X
        else:
            weights = self.mod[view_name].obsm[f"{self.spatial_key}_connectivities"].T
            self.mod[view_name].varm['weighted'] = (weights @ self.mod[view_name].X).T

    def get_weighted_matrix(self, view_name, predictors=None):
        if predictors is None:
            predictors = self.mod[view_name].var_names

        if self.enforce_obs:
            return self.mod[view_name][:, predictors].layers['weighted']
        else:
            return self.mod[view_name][:, predictors].varm['weighted'].T


    @d.dedent
    def __call__(self,
                 model: SingleViewModel,
                 bypass_intra: bool = False,
                 predict_self: bool = False,
                 maskby = None,
                 k_cv: int = 10,
                 alphas = np.array([0.1, 1, 10]),
                 seed: int = V.seed,
                 inplace: bool = V.inplace,
                 verbose: bool = V.verbose,
                 **kwargs
                 ):
        """
        A Multi-view Learning for dissecting Spatial Transcriptomics data (MISTy) model.

        Parameters
        ----------
        model : `str`, optional (default: 'rf')
            Single-view model of class SingleViewModel. Default options are RandomForestModel, LinearModel, and RobustLinearModel
            available via ``liana.method.sp._misty._single_view_models``.
        bypass_intra : `bool`, optional (default: False)
            Whether to bypass modeling the intraview via leave-one-feature-out (LOFO).
            In other words, whether to bypass modelling each target by LOFO within the same spots.
        predict_self : `bool`, optional (default: False)
            Whether to predict self-interactions. These are determined purely by the feature names.
        maskby : `str`, optional (default: None)
            Column in the .obs attribute used to group or mask observations in the intra-view
            If None, all cells are considered as one group.
        k_cv : `int`, optional (default: 10)
            Number of folds for cross-validation used in the multi-view model,
            and single-view models if model is 'linear'.
        alphas : `list`, optional (default: [0.1, 1, 10])
            List of alpha values used to choose from, that control the strength of the ridge regression,
            used for the multi-view part of the model. Only used if there are more than 2 views being modeled (including intra).
        %(seed)s
        %(inplace)s
        %(verbose)s
        **kwargs : `dict`
            Keyword arguments passed to the Regressors. Note that random_state is already set via ``seed``.

        Returns
        -------
        If inplace is True, the results are written to the `.uns` attribute of the object.
        Otherwise two DataFrames are returned, one for target metrics and one for importances.

        """
        model = model(seed, **kwargs)
        view_str = list(self.view_names)
        obs_masks = _create_obs_masks(self.mod['intra'], maskby)

        if bypass_intra:
            view_str.remove('intra')
        intra = self.mod['intra']

        targets_list, importances_list = [], []
        intra_features = intra.var_names.to_list()
        progress_bar = tqdm(intra_features, disable=not verbose)

        for target in (progress_bar):
            for intra_group in obs_masks.keys():
                msk = obs_masks[intra_group]
                importance_dict = {}
                if verbose:
                    d = f"Now learning: {target}" + \
                        (f" masked by {intra_group}" if intra_group is not None else "")
                    progress_bar.set_description(d)

                predictors_nonself, insert_index = _get_nonself(target, intra_features)
                y = intra[msk, target].X.toarray().flatten()
                X = intra[msk, predictors_nonself].X.toarray()

                if not bypass_intra:
                    model.fit(y=y,
                              X=X,
                              predictors=predictors_nonself,
                              k_cv=k_cv,
                              )
                    predictions_intra, importance_dict["intra"] = \
                        model.predictions, model.importances

                    if insert_index is not None and predict_self:
                        # add self-interactions as nan
                        importance_dict["intra"][target] = np.nan

                # store the predictions for each view to construct predictor matrix for meta model
                predictions_list = []

                if not bypass_intra:
                    predictions_list.append(predictions_intra)

                # model the juxta and paraview (if applicable)
                for view_name in [v for v in view_str if v != "intra"]:
                    extra = self.mod[view_name]

                    extra_features = extra.var_names.to_list()
                    _predictors, _ =  _get_nonself(target, extra_features) if not predict_self else (extra_features, None)

                    X = self.get_weighted_matrix(view_name, _predictors).toarray()
                    X = X[msk, :]
                    model.fit(y=y,
                              X=X,
                              predictors=_predictors,
                              k_cv=k_cv,
                              )
                    predictions_extra, importance_dict[view_name] = \
                        model.predictions, model.importances

                    predictions_list.append(predictions_extra)

                target_metrics = _multi_model(y,
                                              np.column_stack(predictions_list),
                                              intra_group,
                                              bypass_intra,
                                              view_str,
                                              target,
                                              k_cv,
                                              alphas,
                                              seed
                                              )
                targets_list.append(target_metrics)

                importances_df = _format_importances(target=target,
                                                     intra_group=intra_group,
                                                     importance_dict=importance_dict
                                                     )
                importances_list.append(importances_df)

        target_metrics, importances = _concat_dataframes(targets_list,
                                                         importances_list,
                                                         view_str)

        if inplace:
            self.uns[K.target_metrics] = target_metrics
            self.uns[K.interactions] = importances
        else:
            return target_metrics, importances


def _create_dict(**kwargs):
    return {k: v for k, v in kwargs.items() if v is not None}

def _format_targets(target, intra_group, view_str, intra_r2, multi_r2, coefs):
    d = _create_dict(target=target,
                     intra_group=intra_group,
                     intra_R2=intra_r2,
                     multi_R2=multi_r2,
                     gain_R2=multi_r2 - intra_r2,
                     )

    target_df = pd.DataFrame(d, index=[0])
    target_df[view_str] = coefs

    return target_df


def _format_importances(target, intra_group, importance_dict):

    importances_df = pd.DataFrame(importance_dict).reset_index().rename(columns={'index': 'predictor'})
    importances_df[['target', 'intra_group']] = target, intra_group

    return importances_df


def _concat_dataframes(targets_list, importances_list, view_str):
    target_metrics = pd.concat(targets_list, axis=0, ignore_index=True)
    importances = pd.concat(importances_list, axis=0, ignore_index=True)
    importances = pd.melt(importances,
                          id_vars=["target", "predictor", "intra_group"],
                          value_vars=view_str,
                          var_name="view",
                          value_name="importances"
                          )

    # drop intra and extra group columns if they are all None
    importances = importances.dropna(axis=1, how='all')
    importances = importances.dropna(axis=0)

    return target_metrics, importances


def _multi_model(y, predictions, intra_group, bypass_intra, view_str, target, k_cv, alphas, seed):
    n_views = len(view_str)

    if (predictions.shape[0] < k_cv) or (y.var() == 0.0):
        if predictions.shape[0] < k_cv:
            warning_message = (f"Number of samples is less than k_cv, {target} metrics set to NaN")
        else:
            warning_message = (f"Variance of '{target}' is 0.0, metrics set to NaN")

        _logg(warning_message, verbose=True, level='warn')
        return _format_targets(target,
                               intra_group,
                               view_str,
                               np.nan,
                               np.nan,
                               np.repeat(np.nan, n_views)
                               )

    kf = KFold(n_splits=k_cv, shuffle=True, random_state=seed)
    R2_vec_intra, R2_vec_multi = np.zeros(k_cv), np.zeros(k_cv)
    coef_mtx = np.zeros((k_cv, n_views))

    model = RidgeCV(alphas=alphas) if n_views > 2 else LinearRegression()

    for cv_idx, (train_index, test_index) in enumerate(kf.split(predictions)):
        multi_model = model.fit(X=predictions[train_index], y=y[train_index])
        R2_vec_multi[cv_idx] = multi_model.score(X=predictions[test_index], y=y[test_index])
        coef_mtx[cv_idx, :] = multi_model.coef_

        if not bypass_intra:
            pred_train = predictions[train_index, 0].reshape(-1, 1)
            pred_test = predictions[test_index, 0].reshape(-1, 1)

            intra_model = model.fit(X=pred_train, y=y[train_index])
            R2_vec_intra[cv_idx] = intra_model.score(X=pred_test, y=y[test_index])

    # format R2s
    intra_r2 = R2_vec_intra.mean().clip(min=0) if not bypass_intra else 0
    multi_r2 = R2_vec_multi.mean().clip(min=0)

    # format coefficients
    coefs = coef_mtx.mean(axis=0).clip(min=0)
    coefs = coefs / coefs.sum()

    # format metrics to a dataframe
    target_metrics = _format_targets(target,
                                     intra_group,
                                     view_str,
                                     intra_r2,
                                     multi_r2,
                                     coefs
                                     )

    return target_metrics

def _get_nonself(target, predictors):
    if target in predictors:
        insert_idx = np.where(np.array(predictors) == target)[0][0]
        predictors_subset = predictors.copy()
        predictors_subset.pop(insert_idx)
    else:
        predictors_subset = predictors
        insert_idx = None
    return predictors_subset, insert_idx


def _create_obs_masks(intra, maskby):
    obs_masks = {}
    # if maskby is a column of only boleans take it as is
    if maskby is None:
        obs_masks[None] = np.ones(intra.shape[0], dtype=bool)
    elif intra.obs[maskby].dtype == bool:
        obs_masks[None] = intra.obs[maskby]
    # else if maskby is column of strings convert to categorical
    elif intra.obs[maskby].dtype == 'category':
        for intra_group in intra.obs[maskby].cat.categories:
            obs_masks[intra_group] = intra.obs[maskby] == intra_group
    else:
        raise ValueError(f"maskby column {maskby} must be a column of booleans or categorical")

    return obs_masks



import pathlib
from scanpy.datasets import pbmc68k_reduced
from pandas.testing import assert_frame_equal
from pandas import read_csv
import numpy as np
from pandas import DataFrame
from itertools import product

from liana.method.sc._liana_pipe import liana_pipe, _expm1_base, _calc_log2fc
from liana._constants import DefaultValues as V

test_path = pathlib.Path(__file__).parent
adata = pbmc68k_reduced()

groupby = 'bulk_labels'

# Test ALL Default parameters
def test_liana_pipe_defaults():
    all_defaults = liana_pipe(adata=adata,
                              groupby=groupby,
                              resource_name=V.resource_name,
                              groupby_pairs=V.groupby_pairs,
                              expr_prop=V.expr_prop,
                              min_cells=V.min_cells,
                              de_method=V.de_method,
                              base=V.logbase,
                              n_perms=V.n_perms,
                              seed=V.seed,
                              verbose=V.seed,
                              supp_columns=[],
                              resource=V.resource,
                              use_raw=V.use_raw,
                              layer=V.layer,
                              n_jobs=1,
                              interactions=V.interactions,
                              )

    assert 'prop_min' in all_defaults.columns

    exp_defaults = read_csv(test_path.joinpath("data", "all_defaults.csv"), index_col=0)
    exp_defaults.index = all_defaults.index
    assert_frame_equal(all_defaults, exp_defaults, check_dtype=False,
                       check_exact=False, check_index_type=False, rtol=1e-3)


# Test NOT Default parameters
def test_liana_pipe_not_defaults():
    not_defaults = liana_pipe(adata=adata,
                              groupby=groupby,
                              resource_name=V.resource_name,
                              expr_prop=0.2,
                              groupby_pairs=V.groupby_pairs,
                              min_cells=V.min_cells,
                              de_method='wilcoxon',
                              base=V.logbase,
                              n_perms=V.n_perms,
                              seed=V.seed,
                              verbose=V.verbose,
                              supp_columns=['ligand_pvals', 'receptor_pvals'],
                              resource=V.resource,
                              use_raw=V.use_raw,
                              layer=V.layer,
                              return_all_lrs=True,
                              n_jobs=1,
                              interactions=V.interactions,
                              )

    assert all(np.isin(['lrs_to_keep'], not_defaults.columns))
    assert all(np.isin(['ligand_pvals', 'receptor_pvals'], not_defaults.columns))

    exp_defaults = read_csv(test_path.joinpath("data/not_defaults.csv"), index_col=0)
    exp_defaults.index = not_defaults.index
    assert_frame_equal(not_defaults, exp_defaults, check_dtype=False,
                       check_index_type=False, check_exact=False, rtol=1e-3)



def test_liana_pipe_subset():
    cts = ['CD34+', 'Dendritic', 'CD56 NK', 'CD19+ B']
    groupby_pairs = list(product(cts, cts))
    groupby_pairs = DataFrame(groupby_pairs, columns=['source', 'target'])
    groupby_pairs = groupby_pairs[groupby_pairs['source'] == 'Dendritic']

    subset = liana_pipe(adata=adata,
                        groupby=groupby,
                        resource_name=V.resource_name,
                        expr_prop=0.05,
                        groupby_pairs=groupby_pairs,
                        min_cells=V.min_cells,
                        de_method=V.de_method,
                        base=V.logbase,
                        n_perms=V.n_perms,
                        seed=V.seed,
                        verbose=V.verbose,
                        resource=V.resource,
                        use_raw=V.use_raw,
                        layer=V.layer,
                        n_jobs=1,
                        interactions=V.interactions,
                        )

    subset.shape == (46, 21)


def test_expm1_fun():
    expm1_mat = _expm1_base(V.logbase, adata.raw.X.data)
    np.testing.assert_almost_equal(np.sum(expm1_mat), 1057526.4, decimal=1)


def test_calc_log2fc():
    adata.layers['normcounts'] = adata.raw.X.copy()
    adata.layers['normcounts'].data = _expm1_base(V.logbase, adata.raw.X.data)
    adata.obs['@label'] = adata.obs.bulk_labels
    np.testing.assert_almost_equal(np.mean(_calc_log2fc(adata, "Dendritic")), -0.123781264)



from liana.method.sc._Method import Method, MethodMeta
from numpy import mean

def _logfc_score(x):
    mean_logfc = mean((x['ligand_logfc'], x['receptor_logfc']), axis=0)
    return None, mean_logfc


# Initialize CPDB Meta
_logfc = MethodMeta(method_name="log2FC",
                    complex_cols=['ligand_means', 'receptor_means'],
                    add_cols=['ligand_logfc', 'receptor_logfc'],
                    fun=_logfc_score,
                    magnitude=None,
                    magnitude_ascending=None,
                    specificity='lr_logfc',
                    specificity_ascending=False,
                    permute=False,
                    reference='Dimitrov, D., Türei, D., Garrido-Rodriguez, M., Burmedi, P.L., '
                              'Nagai, J.S., Boys, C., Ramirez Flores, R.O., Kim, H., Szalai, B., '
                              'Costa, I.G. and Valdeolivas, A., 2022. Comparison of methods and '
                              'resources for cell-cell communication inference from single-cell '
                              'RNA-Seq data. Nature Communications, 13(1), pp.1-13. '
                    )

logfc = Method(_method=_logfc)



import liana.plotting as pl
from liana.testing import generate_toy_spatial
from liana.testing import _sample_target_metrics, _sample_interactions

adata = generate_toy_spatial()
adata.uns['target_metrics'] = _sample_target_metrics()
adata.uns['interactions'] = _sample_interactions()
adata.view_names = ['intra', 'extra']

# test aggregate
target_metrics = adata.uns['target_metrics'].copy()
target_metrics = target_metrics._append(target_metrics)
target_metrics['group'] = ['a'] * 3 + ['b'] * 3

interactions = adata.uns['interactions'].copy()
interactions = interactions._append(interactions)
interactions['group'] = ['a'] * 9 + ['b'] * 9

def test_target_contributions_plot():
    pl.contributions(misty=adata)
    pl.contributions(misty=adata, return_fig=False)


def test_target_metrics_plot():
    pl.target_metrics(misty=adata, stat='gain_R2')
    pl.target_metrics(misty=adata, stat='gain_R2', top_n=1, return_fig=False)
    pl.target_metrics(misty=adata, stat='gain_R2', filter_fun = lambda x: x['multi_R2'] > 0.5)


def test_interactions_plot():
    pl.interactions(misty=adata, top_n=3, view='extra', key=abs, ascending=False)
    plot_data = pl.interactions(interactions=interactions, view='extra', filter_fun=lambda x: x['group']=='b').data
    assert plot_data.shape[0] == 3


def test_target_metrics_aggregate():
    pl.target_metrics(target_metrics=target_metrics, stat='gain_R2', aggregate_fun='mean')

def test_contributions_aggregate():
    pl.contributions(target_metrics=target_metrics, view_names=['intra', 'extra'], aggregate_fun='median')

def test_interactions_aggregate():
    pl.interactions(interactions=interactions, view='intra', aggregate_fun='sum')



from sklearn.ensemble import RandomForestRegressor
from sklearn.linear_model import LinearRegression
import statsmodels.api as sm
import numpy as np
from sklearn.model_selection import KFold, cross_val_predict

class SingleViewModel:
    """
    Base class for single view models. Subclasses should implement the fit method.
    """
    def __init__(self, seed, **kwargs):
        self.seed = seed
        self.kwargs = kwargs  # Store kwargs to be used in fit method
        self.model = None
        self.predictions = None
        self.importances = None

    def fit(self, y, X, predictors, k_cv=None):
        """
        Fit the model to the data and store the predictions and importances.

        Parameters
        ----------
        y : np.ndarray
            Target variable
        X : np.ndarray
            Feature matrix
        predictors : list
            List of feature names
        k_cv : int
            Number of cross-validation folds. If None, no cross-validation is performed.
        """
        raise NotImplementedError("This method should be implemented by subclasses")

    def _k_fold_predict(self, y, X, k_cv, fit_method):
        predictions = np.zeros_like(y)
        kf = KFold(n_splits=k_cv, random_state=self.seed, shuffle=True)
        for train_index, test_index in kf.split(X):
            X_train, X_test = X[train_index], X[test_index]
            y_train = y[train_index]
            model = fit_method(y_train, X_train)
            y_pred = model.predict(X_test)
            predictions[test_index] = y_pred.flatten()
        return predictions

class RandomForestModel(SingleViewModel):
    """
    Random forest model (from sklearn) using out-of-bag predictions for feature importances.
    """

    def fit(self, y, X, predictors, k_cv=None):
        self.model = RandomForestRegressor(oob_score=True, random_state=self.seed, **self.kwargs)
        self.model.fit(X, y)
        self.predictions = self.model.oob_prediction_
        self.importances = dict(zip(predictors, self.model.feature_importances_))

class LinearModel(SingleViewModel):
    """
    Linear regression model using statsmodels.OLS for feature importances, and cross_val_predict with sklearn LinearRegression for predictions.
    **kwargs are passed to statsmodels.OLS.
    """
    def fit(self, y, X, predictors, k_cv):
        # pop n_jobs if it exists
        n_jobs = self.kwargs.pop('n_jobs', -1)
        model = LinearRegression(n_jobs=1)
        self.predictions = cross_val_predict(model,
                                             X, y,
                                             cv=KFold(n_splits=k_cv,
                                                      random_state = self.seed,
                                                      shuffle=True),
                                             n_jobs=n_jobs
                                             )
        X = sm.add_constant(X)
        model_full = sm.OLS(y, X, **self.kwargs).fit()
        self.importances = dict(zip(predictors, model_full.tvalues[1:]))

    def _fit_ols(self, y, X):
        return LinearRegression(**self.kwargs).fit(y=y, X=X)

class RobustLinearModel(SingleViewModel):
    """
    Robust linear regression model using statsmodels.RLM.
    """
    def fit(self, y, X, predictors, k_cv):
        X = sm.add_constant(X)
        self.predictions = self._k_fold_predict(y, X, k_cv, self._fit_robust)
        model_full = sm.RLM(y, X, **self.kwargs).fit()
        self.importances = dict(zip(predictors, model_full.tvalues[1:]))

    def _fit_robust(self, y, X):
        return sm.RLM(y, X, **self.kwargs).fit()



from .select_resource import select_resource
from .select_resource import show_resources
from ._resource_utils import generate_lr_geneset
from ._reassemble_complexes import filter_reassemble_complexes, explode_complexes
from .get_metalinks import get_metalinks, get_metalinks_values, describe_metalinks
from ._orthology import translate_column, translate_resource, get_hcop_orthologs



__version__ = '1.3.0'

from liana import method as mt, plotting as pl, resource as rs, multi as mu, utils as ut, testing

# done after everything has been imported (adapted from scanpy)
import sys
from scanpy._utils import annotate_doc_types

sys.modules.update({f'{__name__}.{m}': globals()[m] for m in ['mt', 'pl', 'rs', 'ut']})
annotate_doc_types(sys.modules[__name__], 'liana')

del sys, annotate_doc_types



import numpy as np
from liana.testing._sample_anndata import generate_toy_adata
from liana.multi import nmf, estimate_elbow

adata = generate_toy_adata()
adata.X = np.abs(adata.X)


def test_run_nmf():
    W, H, _, _ = nmf(adata, n_components=2, inplace=False)

    assert W.shape == (adata.n_obs, 2)
    assert H.shape == (adata.n_vars, 2)

    nmf(adata, n_components=None, inplace=True, random_state=0, max_iter=20)
    assert 'NMF_W' in adata.obsm
    assert 'NMF_H' in adata.varm
    assert adata.obsm['NMF_W'].shape == (adata.n_obs, 4)
    assert adata.varm['NMF_H'].shape == (adata.n_vars, 4)
    assert adata.uns['nmf_errors'].shape == (10, 2)
    assert adata.uns['nmf_rank'] == 4

def test_estimate_elbow():
    errors, rank = estimate_elbow(adata.X, k_range=range(1, 10), random_state=0, max_iter=20)
    assert rank == 4
    assert errors.shape == (9, 2)
    assert errors['k'].tolist() == list(range(1, 10))
    np.testing.assert_almost_equal(errors['error'].mean(), 0.3640689)


def test_run_nmf_df():
    df = adata.to_df()
    W, H, errors, n_components = nmf(df=df, n_components=2, inplace=True, random_state=0, max_iter=20)

    assert W.shape == (adata.n_obs, 2)
    assert H.shape == (adata.n_vars, 2)
    assert n_components == 2
    assert errors is None



import numpy as np
from scipy.sparse import lil_matrix, csr_matrix
from scipy.interpolate import griddata
from anndata import AnnData

from liana.method._pipe_utils._pre import _choose_mtx_rep
from liana._docs import d
from liana._constants import DefaultValues as V


@d.dedent
def interpolate_adata(target: AnnData,
                      reference: AnnData,
                      spatial_key: str,
                      layer=V.layer,
                      use_raw=V.use_raw,
                      method: str = 'linear',
                      fill_value: float = 0,
                      verbose=V.verbose) -> AnnData:
    """
    Interpolates spatial data from a target AnnData object to a reference AnnData object based on spatial coordinates.
    The function creates a new AnnData object where the `.X` attribute is filled with interpolated data using the specified method.

    Parameters
    -----------
    target
        The AnnData object to be interpolated.
    reference
        The AnnData object to be used as reference.
    %(spatial_key)s
    %(layer)s
    %(use_raw)s
    method
        Interpolation method. See `scipy.interpolate.griddata` for more information.
    fill_value
        Value to fill in for points outside of the convex hull of the input points.
    %(verbose)s

    Returns
    --------
    AnnData: A new AnnData object with the same metadata as the reference but with interpolated spatial data in `.X`.
    """
    target_coords = target.obsm[spatial_key]
    reference_coords = reference.obsm[spatial_key]

    X = lil_matrix((reference.shape[0], target.shape[1]), dtype=np.float32)

    ad = AnnData(X=X,
                 uns=reference.uns,
                 obs=reference.obs,
                 obsm=reference.obsm,
                 obsp=reference.obsp,
                 var=target.var,
                 varm=target.varm
                 )

    values = _choose_mtx_rep(adata=target, use_raw=use_raw, layer=layer, verbose=verbose).A

    ad.X = csr_matrix(
        griddata(points=target_coords,
                 xi=reference_coords,
                 values=values,
                 method=method,
                 fill_value=fill_value
                 )
        )

    return ad



from pandas import read_csv
from numpy import unique
import pathlib
from pandas import DataFrame

from liana._logging import _logg
from liana._constants import DefaultValues as V

def select_resource(resource_name: str = V.resource_name) -> DataFrame:
    """
    Read resource of choice from the pre-generated resources in LIANA.

    Parameters
    ----------
    resource_name
        Name of the resource to be loaded and use for ligand-receptor inference.

    Returns
    -------
    A dataframe with ``['ligand', 'receptor']`` columns

    """
    resource_name = resource_name.lower()

    resource_path = pathlib.Path(__file__).parent.joinpath("omni_resource.csv")

    resource = read_csv(resource_path, index_col=False)

    if resource_name not in resource['resource'].unique():
        raise ValueError(f"Resource {resource_name} not found. "
                         f"Please choose from {resource['resource'].unique()}")

    resource = resource[resource['resource'] == resource_name]

    resource = resource[['source_genesymbol', 'target_genesymbol']]
    resource = resource.rename(columns={'source_genesymbol': 'ligand',
                                        'target_genesymbol': 'receptor'}
                               )

    return resource


def show_resources():
    """
    Show available resources.

    Returns
    -------
    A list of resource names available via ``liana.resource.select_resource``

    """
    resource_path = pathlib.Path(__file__).parent.joinpath("omni_resource.csv")
    resource = read_csv(resource_path, index_col=False)
    return list(unique(resource['resource']))


def _handle_resource(interactions=None, resource=None, resource_name=None, x_name='ligand', y_name='receptor', verbose=True):
    if interactions is None:
        if resource is None:
            if resource_name is None:
                raise ValueError("If 'interactions' and 'resource' are both None, 'resource_name' must be provided.")
            else:
                _logg(f"Using resource `{resource_name}`.", verbose=verbose)
                resource = select_resource(resource_name)
        else:
            if verbose:
                print("Using provided `resource`.")
            if not isinstance(resource, DataFrame) or x_name not in resource.columns or y_name not in resource.columns:
                raise ValueError("If 'interactions' is None, 'resource' must be a valid DataFrame "
                                 "with columns '{}' and '{}'.".format(x_name, y_name))
            resource = resource.copy()
            resource = resource.dropna(subset=[x_name, y_name]).drop_duplicates()
            resource.index = range(len(resource))
            resource.index.name = None
    else:
        _logg("Using provided `interactions`.", verbose=verbose)
        if not isinstance(interactions, list) or any(len(item) != 2 for item in interactions):
            raise ValueError("'interactions' should be a list of tuples in the format [(x1, y1), (x2, y2), ...].")
        interactions = set(interactions)
        resource = DataFrame(interactions, columns=[x_name, y_name])

    return resource



import numpy as np
import pathlib
from scanpy.datasets import pbmc68k_reduced
from pandas import read_csv

from liana.method._pipe_utils._get_mean_perms import _get_means_perms, _get_positions
from liana.method.sc._liana_pipe import _trimean
from liana.method.sc._cellphonedb import _mean

test_path = pathlib.Path(__file__).parent

adata = pbmc68k_reduced()
adata.X = adata.raw.X
adata.obs['@label'] = adata.obs.bulk_labels

all_defaults = read_csv(test_path.joinpath("data/all_defaults.csv"), index_col=0)


def test_perms():
    perms = _get_means_perms(adata=adata,
                             norm_factor=None,
                             agg_fun=_mean,
                             n_perms=100,
                             seed=1337,
                             n_jobs=2,
                             verbose=False)

    assert perms.shape == (100, 10, 765)

    desired = np.array([45615.15418553, 45737.95729483, 45575.47318892, 45559.41832494,
                        45542.32316456, 45593.50440302, 45591.03955124, 45561.79108855,
                        45698.86540851, 45543.95444739])
    expected = np.sum(np.sum(perms, axis=0), axis=1)

    np.testing.assert_almost_equal(desired, expected, decimal=3)


def test_positions():
    ligand_pos, receptor_pos, labels_pos = _get_positions(adata, all_defaults)

    assert ligand_pos['MIF'] == 740
    assert receptor_pos['CD4'] == 465
    assert labels_pos['Dendritic'] == 9


def test_cellchat_perms():
    mat_max = adata.X.max()

    perms = _get_means_perms(adata=adata,
                             norm_factor=None,
                             agg_fun=_trimean,
                             n_perms=100,
                             seed=1337,
                             n_jobs=1,
                             verbose=False
                             )

    assert perms.shape == (100, 10, 765)

    desired = np.array([33840.83, 36332.442, 34569.577, 33819.275,
                        33809.956, 33785.234, 33844.524, 34986.043,
                        34304.404, 33644.323])
    expected = perms.sum(axis=0).sum(axis=1)

    assert np.testing.assert_almost_equal(desired, expected, decimal=3) is None

    perms = _get_means_perms(adata=adata,
                             norm_factor=mat_max,
                             agg_fun=_trimean,
                             n_perms=100,
                             seed=1337,
                             n_jobs=1,
                             verbose=False
                             )
    desired = np.array([5215.107487, 5599.082231, 5327.412358,
                        5211.785598, 5210.349528, 5206.53966,
                        5215.676758, 5391.592763, 5286.547464,
                        5184.824284])
    expected = perms.sum(axis=0).sum(axis=1)

    assert np.testing.assert_almost_equal(desired, expected, decimal=6) is None



import os
import sqlite3
import pandas as pd
from typing import Optional, List
from liana._logging import _logg, _check_if_installed

def _download_metalinksdb(verbose=True):
    """
    Ensures the Metalinksdb is downloaded and available for use.
    If the Metalinks database is not present in the current working directory, it downloads it.

    Returns
    -------
    str
        The path to the downloaded database file.
    """
    requests = _check_if_installed("requests")

    METALINKS_URL = "https://figshare.com/ndownloader/files/47567597"

    # Define the local filename to save the downloaded database
    db_file_name = 'metalinksdb.db'
    db_path = os.path.join(os.getcwd(), db_file_name)

    # Check if the database file already exists
    if not os.path.exists(db_path):
        _logg("Downloading database...", verbose=verbose)
        response = requests.get(METALINKS_URL)
        with open(db_path, 'wb') as f:
            f.write(response.content)
        _logg(f"Database downloaded and saved to {db_path}.", verbose=verbose)

    return db_path

def _format_clauses(input_data, column_name, table_ref, where_clauses):
    if input_data:
        formatted_str = ", ".join([f"'{i}'" for i in input_data])
        where_clauses.append(f"{table_ref}.{column_name} IN ({formatted_str})")

def get_metalinks(db_path: Optional[str] = None,
                  types: Optional[List[str]] = None,
                  cell_location: Optional[List[str]] = None,
                  tissue_location: Optional[List[str]] = None,
                  biospecimen_location: Optional[List[str]] = None,
                  disease: Optional[List[str]] = None,
                  pathway: Optional[List[str]] = None,
                  hmdb_ids: Optional[List[str]] = None,
                  uniprot_ids: Optional[List[str]] = None,
                  source: Optional[List[str]] = None
                  ):
    """
    Fetches edges of metabolite-proteins with specified annotations, applying filters if they are not None.
    Allows filtering by lists of hmdb and uniprot IDs and avoids duplicate column names, and returns the results as a pandas DataFrame.
    Filters are applied using INNER JOINs and WHERE clauses - i.e. the results are the intersection of the filters.

    Parameters
    ----------
    db_path
        Path to the SQLite database file. If None, the database will be downloaded to the current working directory.
    types
        Desired edge types. Options are: ['lr', 'pd'], where 'lr' stands for 'ligand-receptor' and 'pd' stands for 'production-degradation'.
    cell_location
        Desired metabolite cell locations.
    tissue_location
        Desired metabolite tissue locations.
    biospecimen_location
        Desired metabolite biospecimen locations.
    disease
        Desired metabolite diseases.
    pathway
        Desired metabolite pathways.
    hmdb_ids
        Desired HMDB IDs.
    uniprot_ids
        Desired UniProt IDs.

    Returns
    ----------

    A pandas DataFrame containing the query results without the source column.
    """

    if db_path is None:
        db_path = _download_metalinksdb()
    conn = sqlite3.connect(db_path)

    # Adjusted SELECT statement to exclude the source column
    base_query = """
    SELECT DISTINCT e.hmdb as hmdb,
                e.uniprot AS uniprot,
                p.gene_symbol as gene_symbol,
                m.metabolite AS metabolite,
                e.mor as mor,
                e.transport_direction as transport_direction,
                e.type AS type,
                e.source AS source
    FROM edges e
    LEFT JOIN metabolites m ON e.hmdb = m.hmdb
    LEFT JOIN proteins p ON e.uniprot = p.uniprot
    """

    def _to_list(x):
        if isinstance(x, str):
            return [x]
        return x

    cell_location = _to_list(cell_location)
    tissue_location = _to_list(tissue_location)
    biospecimen_location = _to_list(biospecimen_location)
    disease = _to_list(disease)
    pathway = _to_list(pathway)
    hmdb_ids = _to_list(hmdb_ids)
    uniprot_ids = _to_list(uniprot_ids)
    types = _to_list(types)

    annotations_filters = {
        'cell_location': cell_location,
        'tissue_location': tissue_location,
        'biospecimen_location': biospecimen_location,
        'disease': disease,
        'pathway': pathway
    }

    join_clauses = []
    where_clauses = []
    for annotation_table, values in annotations_filters.items():
        if values is not None:
            join_clause = f"INNER JOIN {annotation_table} ON m.hmdb = {annotation_table}.hmdb"
            join_clauses.append(join_clause)

            values_str = ", ".join([f"'{value}'" for value in values])
            where_clause = f"{annotation_table}.{annotation_table} IN ({values_str})"
            where_clauses.append(where_clause)

    _format_clauses(types, "type", "e", where_clauses)
    _format_clauses(hmdb_ids, "hmdb", "m", where_clauses)
    _format_clauses(uniprot_ids, "uniprot", "p", where_clauses)
    _format_clauses(source, "source", "e", where_clauses)

    full_query = base_query
    if join_clauses:
        full_query += " " + " ".join(join_clauses)
    if where_clauses:
        full_query += " WHERE " + " AND ".join(where_clauses)

    df = pd.read_sql_query(full_query, conn)
    conn.close()

    return df


def get_metalinks_values(table_name, column_name, db_path: Optional[str] = None):
    """
    Fetches distinct values from a specified column in a specified table.

    Parameters
    ----------
    db_path : str
        Path to the SQLite database file. If None, the database will be downloaded to the current working directory.
    table_name : str
        Name of the table from which to fetch distinct values.
    column_name : str
        Name of the column from which to fetch distinct values.

    Returns
    -------
    list
        A list of distinct values from the specified column.
    """
    if db_path is None:
        db_path = _download_metalinksdb()
    conn = sqlite3.connect(db_path)
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    query = f"SELECT DISTINCT {column_name} FROM {table_name};"
    cursor.execute(query)
    distinct_values = cursor.fetchall()
    conn.close()
    return [value[0] for value in distinct_values]


def describe_metalinks(db_path: Optional[str] = None, return_output: bool = False):
    """
    Prints the schema information and foreign key details for all tables in the specified SQLite database.

    Parameters
    ----------
    db_path : str
        Path to the SQLite database file. If None, the database will be downloaded to the current working directory.
    """
    if db_path is None:
        db_path = _download_metalinksdb()
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
    tables = cursor.fetchall()

    output = ""
    for table in tables:
        table_name = table[0]
        output += f"Schema of table: {table_name}\n{'=' * len(f'Schema of table: {table_name}')}\n"

        cursor.execute(f"PRAGMA table_info({table_name});")
        schema_info = cursor.fetchall()
        for column in schema_info:
            cid, name, ctype, _, _, pk = column
            output += f"Column ID: {cid}, Name: {name}, Type: {ctype}, Primary Key: {pk}\n"

        cursor.execute(f"PRAGMA foreign_key_list({table_name});")
        fk_info = cursor.fetchall()
        if fk_info:
            output += "\nForeign Keys:\n"
            for fk in fk_info:
                id, seq, table, from_col, to_col, _, _, _ = fk
                output += f"ID: {id}, Seq: {seq}, Table: {table}, From: {from_col}, To: {to_col}\n"
        else:
            output += "\nNo Foreign Keys.\n"
        output += "-" * 40 + "\n"

    if return_output:
        return output
    else:
        print(output)



from __future__ import annotations

from liana.method.sc._liana_pipe import liana_pipe
from liana._logging import _logg
from liana._docs import d
from liana._constants import Keys as K, DefaultValues as V

import anndata as an
from mudata import MuData
from pandas import DataFrame, concat
from typing import Optional
from tqdm import tqdm
import weakref


class MethodMeta:
    """
    A Class used to store Method Metadata
    """
    # initiate a list to store weak references to all instances
    instances = []

    def __init__(self,
                 method_name: str,
                 complex_cols: list,
                 add_cols: list,
                 fun,
                 magnitude: str | None,
                 magnitude_ascending: bool | None,
                 specificity: str | None,
                 specificity_ascending: bool | None,
                 permute: bool,
                 reference: str
                 ):
        """
        Parameters
        ----------
        method_name
            Name of the Method
        complex_cols
            Columns relevant for protein complexes
        add_cols
            Additional columns required by the method
        fun
            Interaction Scoring function
        magnitude
            Name of the `magnitude` Score (None if not present)
        magnitude_ascending
            Whether to rank `magnitude` in ascending manner (None if not relevant)
        specificity
            Name of the `specificity` Score if Present (None if not present)
        specificity_ascending
            Whether to rank `magnitude` in ascending manner  (None if not relevant)
        permute
            Whether it requires permutations
        reference
            Publication reference in Harvard style
        """
        self.__class__.instances.append(weakref.proxy(self))
        self.method_name = method_name
        self.complex_cols = complex_cols
        self.add_cols = add_cols
        self.fun = fun
        self.magnitude = magnitude
        self.magnitude_ascending = magnitude_ascending
        self.specificity = specificity
        self.specificity_ascending = specificity_ascending
        self.permute = permute
        self.reference = reference

    def describe(self):
        """Briefly described the method"""
        print(
            f"{self.method_name} uses `{self.magnitude}` and `{self.specificity}`"
            f" as measures of expression strength and interaction specificity, respectively"
        )

    def reference(self):
        """Prints out reference in Harvard format"""
        print(self.reference)

    def get_meta(self):
        """Returns method metadata as pandas row"""
        meta = DataFrame([{"Method Name": self.method_name,
                           "Magnitude Score": self.magnitude,
                           "Specificity Score": self.specificity,
                           "Reference": self.reference
                           }])
        return meta

    @d.dedent
    def by_sample(self,
                  adata: an.AnnData | MuData,
                  sample_key: str,
                  key_added: str = K.uns_key,
                  inplace: bool = V.inplace,
                  verbose: bool = V.verbose,
                  **kwargs):
        """
        Run a method by sample.

        Parameters
        ----------
        %(adata)s
        %(sample_key)s
        %(key_added)s
        %(inplace)s
        verbose
            Possible values: False, True, 'full', where 'full' will print the results for each sample,
            and True will only print the sample progress bar. Default is False.
        **kwargs
            keyword arguments to pass to the method

        Returns
        -------
        A pandas DataFrame with the results and a column sample is stored in `adata.uns[key_added]` if `inplace` is True,
        else the DataFrame is returned.

        """

        if sample_key not in adata.obs:
            raise ValueError(f"{sample_key} was not found in `adata.obs`.")

        if not adata.obs[sample_key].dtype.name == "category":
            _logg(f"Converting `{sample_key}` to categorical!", level='warn', verbose=verbose)
            adata.obs[sample_key] = adata.obs[sample_key].astype("category")

        if verbose == 'full':
            verbose = True
            full_verbose = True
        else:
            full_verbose = False

        samples = adata.obs[sample_key].cat.categories

        adata.uns[key_added] = {}

        progress_bar = tqdm(samples, disable=not verbose)
        for sample in (progress_bar):
            if verbose:
                progress_bar.set_description(f"Now running: {sample}")


            temp = adata[adata.obs[sample_key]==sample]
            if temp.isbacked:
                temp = temp.to_memory().copy() # NOTE does to_memory copy?
            else:
                temp = temp.copy()

            sample_res = self.__call__(temp, inplace=False, verbose=full_verbose, **kwargs)

            adata.uns[key_added][sample] = sample_res

        liana_res = concat(adata.uns[key_added]).reset_index(level=1, drop=True).reset_index()
        liana_res = liana_res.rename({"index":sample_key}, axis=1)

        if inplace:
            adata.uns[key_added] = liana_res
        return None if inplace else liana_res


class Method(MethodMeta):
    """
    Ligand-Receptor Method Class
    """
    def __init__(self, _method):
        super().__init__(method_name=_method.method_name,
                         complex_cols=_method.complex_cols,
                         add_cols=_method.add_cols,
                         fun=_method.fun,
                         magnitude=_method.magnitude,
                         magnitude_ascending=_method.magnitude_ascending,
                         specificity=_method.specificity,
                         specificity_ascending=_method.specificity_ascending,
                         permute=_method.permute,
                         reference=_method.reference
                         )
        self._method = _method

    @d.dedent
    def __call__(self,
                 adata: an.AnnData | MuData,
                 groupby: str,
                 resource_name: str = V.resource_name,
                 expr_prop: float = V.expr_prop,
                 min_cells: int = V.min_cells,
                 groupby_pairs: Optional[DataFrame] = V.groupby_pairs,
                 base: float = V.logbase,
                 supp_columns: list = V.supp_columns,
                 return_all_lrs: bool = V.return_all_lrs,
                 key_added: str = K.uns_key,
                 use_raw: Optional[bool] = V.use_raw,
                 layer: Optional[str] = V.layer,
                 de_method: str = V.de_method,
                 n_perms: int = V.n_perms,
                 seed: int = V.seed,
                 n_jobs: int = 1,
                 resource: Optional[DataFrame] = V.resource,
                 interactions: Optional[list] = V.interactions,
                 mdata_kwargs: dict = dict(),
                 inplace: bool = V.inplace,
                 verbose: Optional[bool] = V.verbose,
                 ):
        """
        Run a ligand-receptor method.

        Parameters
        ----------
        %(adata)s
        %(groupby)s
        %(resource_name)s
        %(expr_prop)s
        %(min_cells)s
        %(groupby_pairs)s
        %(base)s
        supp_columns
            Additional columns to be added from any of the methods implemented in liana,
            or any of the columns returned by `scanpy.tl.rank_genes_groups`, each starting with ligand_* or receptor_*.
            For example, `['ligand_pvals', 'receptor_pvals']`. None by default.
        %(return_all_lrs)s
        %(key_added)s
        %(use_raw)s
        %(layer)s
        %(de_method)s
        %(verbose)s
        %(n_perms_sc)s
        %(seed)s
        n_jobs
            Number of jobs to run in parallel.
        %(resource)s
        %(interactions)s
        %(mdata_kwargs)s
        %(inplace)s

        Returns
        -------
            If ``inplace = False``, returns a `DataFrame` with ligand-receptor results
            Otherwise, modifies the ``adata`` object with the following key:
            - :attr:`anndata.AnnData.uns` ``[`key_added`]`` with the aforementioned DataFrame
        """
        if supp_columns is None:
            supp_columns = []

        liana_res = liana_pipe(adata=adata,
                               groupby=groupby,
                               resource_name=resource_name,
                               resource=resource,
                               interactions=interactions,
                               expr_prop=expr_prop,
                               min_cells=min_cells,
                               supp_columns=supp_columns,
                               return_all_lrs=return_all_lrs,
                               groupby_pairs=groupby_pairs,
                               base=base,
                               de_method=de_method,
                               verbose=verbose,
                               _score=self._method,
                               n_perms=n_perms,
                               seed=seed,
                               n_jobs=n_jobs,
                               use_raw=use_raw,
                               layer=layer,
                               mdata_kwargs=mdata_kwargs
                               )
        if inplace:
            adata.uns[key_added] = liana_res
        return None if inplace else liana_res


def _show_methods(methods):
    return concat([method.get_meta() for method in methods])



from __future__ import annotations
from docrep import DocstringProcessor

# Common docstrings
_adata = """\
adata
    Annotated data object."""

_mdata = """\
mdata
    MuData (multimodal) data object."""

_misty = """\
misty
    MistyData object with modelling results
"""

_groupby = """\
groupby
    Key to be used for grouping."""

_n_perms = """\
n_perms
    Number of permutations for the permutation test. If None, no p-values are computed."""

_seed = """\
seed
    Random seed for reproducibility."""

_resource = """\
resource
    A pandas dataframe with [`ligand`, `receptor`] columns.
    If provided will overrule the resource requested via `resource_name`"""

_interactions = """\
interactions
    List of tuples with ligand-receptor pairs `[(ligand, receptor), ...]` to be used for the analysis.
    If passed, it will overrule the resource requested via `resource` and `resource_name`."""

_resource_name = """\
resource_name
    Name of the resource to be used for ligand-receptor inference. See `li.rs.show_resources()` for available resources."""

_sample_key = """\
sample_key
    key in `adata.obs` to use for grouping by sample or context."""

_key_added = """\
key_added
    Key under which the results will be stored in `adata.uns` if `inplace` is True."""

_use_raw = """\
use_raw
    Use raw attribute of adata if present."""

_layer = """\
layer
    Layer in anndata.AnnData.layers to use. If None, use anndata.AnnData.X."""

_mdata_kwargs = """\
mdata_kwargs
    Keyword arguments to be passed to `li.fun.mdata_to_anndata` if `adata` is an instance of `MuData`.
    If an AnnData object is passed, these arguments are ignored."""

_inplace = """\
inplace
    Whether to store results in place, or else to return them."""

_verbose = """\
verbose
    Verbosity flag."""

_lr_sep = """\
lr_sep
    Separator to use when joining ligand and receptor names into interactions.
"""


_source_labels = """\
source_labels
    List of labels to use as `source`, the rest are filtered out.
"""

_target_labels = """\
target_labels
    List of labels to use as `target`, the rest are filtered out.
"""


# Single-cell specific docstrings
_n_perms_sc = """\
n_perms
    Number of permutations for the permutation test. Relevant only for permutation-based methods
    (e.g., `CellPhoneDB`). If `None` is passed, no permutation testing is performed."""

_expr_prop = """\
expr_prop
    Minimum expression proportion for the ligands and receptors (+ their subunits) in the
    corresponding cell identities. Set to 0 to return unfiltered results."""

_min_cells = """\
min_cells
    Minimum cells (per cell identity if grouped by `groupby`) to be considered for downstream analysis."""

_base = """\
base
    Exponent base used to reverse the log-transformation of the matrix. Relevant only for the `logfc` method."""

_return_all_lrs = """\
return_all_lrs
    Bool whether to return all ligand-receptor pairs, or only those that surpass the `expr_prop`
    threshold. Ligand-receptor pairs that do not pass the `expr_prop` threshold will be assigned
    to the *worst* score of the ones that do. `False` by default."""

_de_method = """\
de_method
    Differential expression method. `scanpy.tl.rank_genes_groups` is used to rank genes
    according to 1vsRest. The default method is 't-test'."""

_groupby_pairs = """\
groupby_pairs
    A DataFrame with columns `source` and `target` to be used to subset the possible combinations of interacting cell types.
    If None, all possible combinations are used."""


# multi-condition specific docstrings
_source_key = """\
source_key
    Column name of the sender/source cell types in `liana_res`."""

_target_key = """\
target_key
    Column name of the receiver/target cell types in `liana_res`."""

_ligand_key = """\
ligand_key
    Column name of the ligand in `liana_res`."""

_receptor_key = """\
receptor_key
    Column name of the receptor in `liana_res`."""

_score_key = """\
score_key
    Column name of the score in `liana_res`. If None, the score is inferred from the method."""

_uns_key = """\
uns_key
    Key in `adata.uns` that contains the LIANA results. Default is `'liana_res'`."""


_inverse_fun = """\
inverse_fun
    Function that is applied to the scores before building the views. Default is `lambda x: 1 - x` which is used to invert the scores
    reflect probabilities (e.g. magnitude_rank), i.e. such for which lower values reflect higher relevance.
    This is handled automatically for the scores in liana."""


# Spatial specific docstrings
_spatial_key = """\
spatial_key
    Key in `adata.obsm` that contains the spatial coordinates. Default is `'spatial'`."""

_connectivity_key = """\
connectivity_key
    Key in `adata.obsp` that contains the spatial connectivity matrix. Default is `'spatial_connectivity'`. """

_local_name = """\
local_name
    Name of the local function to use for the analysis. Passing `None` will return only the Global scores."""

_global_name = """\
global_name
    Name or names (list) of the global function(s) to use for the analysis. Passing `None` will not calculate any global scores"""

_positive_only = """\
positive_only
    Whether to mask non-positive interactions."""

_add_categories = """\
add_categories
    Whether to add categories about the local scores."""

_x_mod = """\
x_mod
    Name of the modality to use for the x-axis."""

_y_mod = """\
y_mod
    Name of the modality to use for the y-axis."""

_x_name = """\
x_name
    Name of the x-variable. If passing a `resource` dataframe, this should match the first column."""

_y_name = """\
y_name
    Name of the y-variable. If passing a `resource` dataframe, this should match the second column."""

_mask_negatives = """\
mask_negatives
    Whether to mask negative-negative (low-low) or uncategorized interactions."""


# Plot docstrings
_liana_res = """\
liana_res
    `liana_res` a `DataFrame` in liana's format.
"""

_colour = """\
colour
    `column` in `liana_res` to define the colours of the dots.
"""

_size = """\
size
    `column` in `liana_res` to define the size of the dots.
"""

_top_n = """\
top_n
    ``top_n`` entities to plot.
"""

_orderby = """\
orderby
    If `top_n` is not `None`, order the interactions by this column
"""

_orderby_ascending = """\
orderby_ascending
    If `top_n` is not `None`, specify how to order the interactions
"""

_orderby_absolute = """\
orderby_absolute
    If `top_n` is not `None`, whether to order by the absolute value of the `orderby` column.
"""

_filter_fun = """\
filter_fun
    A function, applied along the columns (axis=1), used to filter the results to be plotted.
"""

_aggregate_fun = """\
aggregate_fun
    A function used to aggregate the results to be plotted.
"""

_ligand_complex = """\
ligand_complex
    `list` of ligand complexes to filter the interactions to be plotted. Defaults to None.
"""

_receptor_complex = """\
receptor_complex
    `list` of receptor complexes to filter the interactions to be plotted. Defaults to None.
"""


_inverse_colour = """\
inverse_colour
    Whether to -log10 the `colour` column for plotting. `False` by default.
"""

_inverse_size = """\
inverse_size
    Whether to -log10 the `size` column for plotting. `False` by default.
"""

_size_range = """\
size_range
    Define size range. Tuple of (min, max) integers.
"""

_cmap = """\
cmap
    Colour map to use for plotting.
"""

_figure_size = """\
figure_size
    Figure x,y size
"""

_return_fig = """\
return_fig
    `bool` whether to return the fig object.
"""


d = DocstringProcessor(
    adata=_adata,
    mdata=_mdata,
    misty=_misty,
    groupby=_groupby,
    groupby_pairs=_groupby_pairs,
    seed=_seed,
    resource=_resource,
    interactions=_interactions,
    resource_name=_resource_name,
    sample_key=_sample_key,
    key_added=_key_added,
    use_raw=_use_raw,
    layer=_layer,
    mdata_kwargs=_mdata_kwargs,
    inplace=_inplace,
    verbose=_verbose,
    lr_sep=_lr_sep,
    n_perms=_n_perms,
    n_perms_sc=_n_perms_sc,
    expr_prop=_expr_prop,
    min_cells=_min_cells,
    base=_base,
    return_all_lrs=_return_all_lrs,
    liana_res=_liana_res,
    de_method=_de_method,
    source_key=_source_key,
    target_key=_target_key,
    ligand_key=_ligand_key,
    receptor_key=_receptor_key,
    score_key=_score_key,
    uns_key=_uns_key,
    inverse_fun=_inverse_fun,
    spatial_key=_spatial_key,
    connectivity_key=_connectivity_key,
    local_name=_local_name,
    global_name=_global_name,
    positive_only=_positive_only,
    add_categories=_add_categories,
    x_mod=_x_mod,
    y_mod=_y_mod,
    x_name=_x_name,
    y_name=_y_name,
    mask_negatives=_mask_negatives,
    colour=_colour,
    size=_size,
    source_labels=_source_labels,
    target_labels=_target_labels,
    top_n=_top_n,
    orderby=_orderby,
    orderby_ascending=_orderby_ascending,
    orderby_absolute=_orderby_absolute,
    filter_fun =_filter_fun,
    aggregate_fun=_aggregate_fun,
    ligand_complex=_ligand_complex,
    receptor_complex=_receptor_complex,
    inverse_colour=_inverse_colour,
    inverse_size=_inverse_size,
    size_range=_size_range,
    cmap=_cmap,
    figure_size=_figure_size,
    return_fig=_return_fig,

)



import numpy as np
from liana._constants import PrimaryColumns as P, DefaultValues as V
from liana._docs import d

@d.dedent
def generate_lr_geneset(resource,
                        net,
                        ligand_key=P.ligand,
                        receptor_key=P.receptor,
                        lr_sep=V.lr_sep,
                        source='source',
                        target='target',
                        weight='weight'
                        ):
    """
    Generate a ligand-receptor gene set from a resource and a network.

    Specifically, it works with weighted bipartite networks, where the weight represents the importance of the genes
    to a given geneset. The function will assign a weight to each ligand-receptor interaction, based on the mean.
    It does so by first assigning a weight to each ligand-receptor subunit, checking for sign coherence and completeness
    of the ligand-receptor complex.

    Parameters
    ----------
    resource:
        A pandas dataframe with [`ligand`, `receptor`] columns.
    net
        Prior knowledge network in bipartite or decoupler format.
    ligand : str, optional
        Name of the ligand column in the resource
    receptor : str, optional
        Name of the receptor column in the resource
    %(lr_sep)s
    source : str, optional
        Name of the source column in the network.
    weight : str, optional
        Name of the weight column in the network. If None, all weights are set to 1.

    Returns
    -------
    Returns ligand-receptor geneset resource as a pandas.DataFrame with the following columns:
    - interaction: ligand-receptor interaction
    - weight: mean weight of the interaction
    - source: source of the interaction
    """
    # TODO: Fix this if else, it's not very elegant
    if weight is None:
        weight = 'weight'
        net[weight] = 1

        drop_weight = True
    else:
        drop_weight = False

    # supp keys
    ligand_weight = ligand_key + '_' + weight
    receptor_weight = receptor_key + '_' + weight
    ligand_source = ligand_key + '_' + source
    receptor_source = receptor_key + '_' + source

    # assign weights to each entity
    ligand_weights = _assign_entity_weights(resource, net, source=source, target=target, entity_key=ligand_key)
    ligand_weights.rename(columns={weight: ligand_weight, source:ligand_source}, inplace=True)
    receptor_weights = _assign_entity_weights(resource, net, source=source, target=target, entity_key=receptor_key)
    receptor_weights.rename(columns={weight: receptor_weight, source: receptor_source}, inplace=True)

    # join weights to the the ligand-receptor resource
    resource = resource.merge(ligand_weights, on=ligand_key, how='inner')
    resource = resource.merge(receptor_weights, on=receptor_key, how='inner')

    # keep only coherent ligand and receptor sources
    resource = resource[resource[ligand_source] == resource[receptor_source]]
    # mean of sign-coherent ligand-receptor weights
    resource.loc[:, weight] = resource.apply(lambda x: _sign_coherent_mean(np.array([x[ligand_weight], x[receptor_weight]])), axis=1)

    # unite ligand-receptor columns
    resource = resource.assign(interaction = lambda x: x[ligand_key] + lr_sep + x[receptor_key])

    # keep only relevant columns
    resource = resource[[ligand_source, 'interaction', weight]].rename(columns={ligand_source: source})

    # drop nan weights
    resource = resource.dropna()

    if drop_weight:
        resource.drop(columns=['weight'], inplace=True)

    return resource

def _assign_entity_weights(resource, net, entity_key='receptor', source='source', target='target', weight='weight'):
    # only keep relevant columns
    net = net[[source, target, weight]]

    # process ligand-receptor resource
    # assign receptor complex as entity
    entity_resource = resource[[entity_key]].drop_duplicates().set_index(entity_key)
    entity_resource['subunit'] =  entity_resource.index
    # explode complexes, keeping the complex as a key
    entity_resource = entity_resource.apply(lambda x: x.str.split('_')).explode(['subunit'])

    # join weights to subunits
    entity_resource = entity_resource.reset_index()
    entity_resource = entity_resource.merge(net, left_on='subunit', right_on=target)

    # check for sign and set consistency
    # count expected subunits separated by _
    entity_resource = entity_resource.assign(subunit_expected = entity_resource[entity_key].str.count('_')+1)
    # count subunits by receptor complex & source
    entity_resource['subunit_count'] = entity_resource.groupby([source, entity_key])[[weight]].transform('count')
    # check if all subunits are present
    entity_resource = entity_resource.assign(subunit_complete = lambda x: x['subunit_expected'] == x['subunit_count'])
    # assign flag to sign-coherent subunits
    entity_resource['sing_coherent'] = entity_resource.groupby([source, entity_key])[[weight]].transform(lambda x: np.all(x > 0) | np.all(x < 0))

    # keep only relevant targets
    entity_resource = entity_resource[entity_resource['subunit_complete']] # keep only complete complexes
    entity_resource = entity_resource[entity_resource['sing_coherent']] # keep only sign-coherent complexes

    # get mean weight per complex & source
    entity_resource = entity_resource.groupby([source, entity_key])[[weight]].mean().reset_index()

    return entity_resource

def _sign_coherent_mean(x):
    if np.all(x > 0) | np.all(x < 0):
        return np.mean(x)
    else:
        return np.nan



import os
import pathlib
import numpy as np
import scanpy as sc
from liana.method.sp._misty._misty_constructs import lrMistyData, genericMistyData
from liana.method.sp._misty._single_view_models import RandomForestModel, LinearModel, RobustLinearModel
from liana.testing._sample_anndata import generate_toy_spatial
from liana.method import MistyData

test_path = pathlib.Path(__file__).parent

adata = sc.read_h5ad(os.path.join(test_path, "data" , "synthetic.h5ad"))
adata = sc.pp.subsample(adata, n_obs=100, copy=True)


def test_misty_para():
    misty = genericMistyData(adata,
                             bandwidth=10,
                             cutoff=0,
                             add_juxta=False,
                             set_diag=False,
                             )
    misty(model=RandomForestModel, bypass_intra=False, seed=42, n_estimators=11)
    assert np.isin(list(misty.uns.keys()), ['target_metrics', 'interactions']).all()
    target_metrics = misty.uns['target_metrics']
    assert np.sum(target_metrics[['intra', 'para']].values, axis=1).sum() == 11.0
    assert target_metrics.shape == (11, 6)

    interactions = misty.uns['interactions']
    assert interactions.shape == (220, 4)
    assert interactions[interactions['target']=='ECM']['importances'].sum().round(8) == 2.0


def test_misty_bypass():
    misty = genericMistyData(adata,
                             bandwidth=10,
                             add_juxta=True,
                             set_diag=True,
                             cutoff=0)
    misty(model=RandomForestModel, alphas=1, bypass_intra=True, seed=42, n_estimators=11)
    assert np.isin(['juxta', 'para'], misty.uns['target_metrics'].columns).all()
    assert ~np.isin(['intra'], misty.uns['target_metrics'].columns).all()
    assert misty.uns['target_metrics'].shape == (11, 6)
    np.testing.assert_almost_equal(misty.uns['target_metrics']['multi_R2'].sum(), 0, decimal=5)

    interactions = misty.uns['interactions']
    assert interactions.shape == (220, 4)
    assert interactions['importances'].sum().round(10) == 22.0
    np.testing.assert_almost_equal(interactions[(interactions['target']=='ligC') &
                                               (interactions['predictor']=='ligA')]['importances'].values,
                                   np.array([0.095, 0.07]), decimal=3)


def test_misty_groups():
    misty = genericMistyData(adata,
                             bandwidth=20,
                             add_juxta=True,
                             set_diag=False,
                             cutoff=0,
                             )
    misty(model=RandomForestModel,
          alphas=1,
          bypass_intra=False,
          seed=42,
          predict_self=True,
          maskby='cell_type',
          n_estimators=11)

    assert misty.uns['target_metrics'].shape==(22, 8)
    perf_actual = (misty.uns['target_metrics'].
     groupby(['intra_group'])['gain_R2'].
     mean().values
    )
    perf_expected = np.array([-0.0124669, -0.0056514])
    np.testing.assert_almost_equal(perf_actual, perf_expected, decimal=2)

    # assert that there are self interactions = var_n * var_n
    interactions = misty.uns['interactions']
    self_interactions = interactions[(interactions['target']==interactions['predictor'])]
    # 11 vars * 4 envs * 3 views = 132; NOTE: However, I drop NAs
    assert self_interactions.shape == (44, 5)
    assert self_interactions[self_interactions['view']=='intra']['importances'].isna().all()


def test_lr_misty():
    adata = generate_toy_spatial()
    misty = lrMistyData(adata, bandwidth=10, set_diag=True, cutoff=0)
    assert misty.shape == (700, 42)

    misty(model=RandomForestModel, n_estimators=10, bypass_intra=True)
    assert misty.uns['target_metrics'].shape == (16, 5)

    interactions = misty.uns['interactions']
    assert interactions.shape == (415, 4)
    cmplxs = interactions[interactions['target'].str.contains('_')]['target'].unique()
    assert np.isin(['CD8A_CD8B', 'CD74_CXCR4'], cmplxs).all()


def test_linear_misty():
    misty = genericMistyData(adata, bandwidth=10, set_diag=False, cutoff=0)
    assert misty.shape == (100, 33)

    misty(model=LinearModel)
    assert misty.uns['target_metrics'].shape == (11, 7)

    assert misty.uns['interactions'].shape == (330, 4)
    actual = misty.uns['interactions']['importances'].values.mean()
    np.testing.assert_almost_equal(actual, 0.5135328101662447, decimal=3)


def test_misty_mask():
    misty = genericMistyData(adata, bandwidth=10, set_diag=False, cutoff=0)
    misty = MistyData(misty)
    assert misty.shape == (100, 33)

    misty.mod['intra'].obs['mask'] = misty.mod['intra'].obs=='A'
    misty(model=LinearModel, maskby='mask')

    assert misty.uns['target_metrics'].shape == (11, 7)
    np.testing.assert_almost_equal(misty.uns['target_metrics']['multi_R2'].mean(), 0.4203699749106394, decimal=3)
    np.testing.assert_almost_equal(misty.uns['target_metrics']['intra_R2'].mean(), 0.4248588250759459, decimal=3)

    assert misty.uns['interactions'].shape == (330, 4)
    np.testing.assert_almost_equal(misty.uns['interactions']['importances'].sum(), 149.30560405771703, decimal=0)


def test_misty_custom():
    adata = generate_toy_spatial()
    # keep first 10 vars
    xdata = adata[:, :10].copy()
    xdata.var.index = 'x' + xdata.var.index

    ydata = adata[:, -10:].copy()
    ydata.var.index = 'y' + ydata.var.index
    intra = adata[:, 25:30].copy()
    misty = MistyData({'intra': intra, 'xdata': xdata, 'ydata': ydata}, verbose=True)
    misty(model=RobustLinearModel, k_cv=25, seed=420)

    misty.uns['interactions'].shape == (120, 4)
    np.testing.assert_almost_equal(misty.uns['interactions']['importances'].max(), 7.427809495362697, decimal=3)
    np.testing.assert_almost_equal(misty.uns['interactions']['importances'].min(), -2.8430222384873396, decimal=3)
    # the data is random
    np.testing.assert_almost_equal(misty.uns['target_metrics']['multi_R2'].mean(), 0, decimal=3)


def test_misty_nonaligned():
    adata = generate_toy_spatial()

    intra = adata[:, :10].copy()
    intra.var.index = 'x' + intra.var.index
    para = adata[:int(adata.n_obs*0.9), -10:].copy()
    para.var.index = 'y' + para.var.index
    # Generate connectivities with shape (para.n_obs, intra.n_obs)
    para.obsm['spatial_connectivities'] = np.ones((para.n_obs, intra.n_obs))
    misty = MistyData({'intra': intra, 'ydata': para},
                      enforce_obs=False, # NOTE: This is the key parameter
                      verbose=True)
    misty(model=LinearModel, k_cv=3)



import numpy as np
import pandas as pd
from liana.method import estimate_metalinks
from liana.testing._sample_anndata import generate_toy_adata

def test_estimate_metalinks():
    adata = generate_toy_adata()
    metabolites = ['A', 'A','B','B']
    resource = pd.DataFrame({'metabolite': np.unique(metabolites),
                             'receptor': ['TNFRSF4', 'ITGB2']})
    pd_net = pd.DataFrame({'metabolite': metabolites,
                          'target': ['RBP7', 'SRM', 'MAD2L2', 'AGTRAP'],
                          'weight': [0.1, 0.2, 0.3, 1]})
    t_net = pd.DataFrame({'metabolite': metabolites,
                         'target': ['HES4', 'TNFRSF4', 'SSU72', 'PARK7'],
                         'weight': [0.1, 0.2, 0.3, 1]})

    mdata = estimate_metalinks(adata, resource,
                               pd_net, t_net,
                               source='metabolite',
                               use_raw=True, verbose=True,
                               min_n=2)
    assert np.isin(['metabolite', 'receptor'], list(mdata.mod.keys())).all()
    assert (mdata.var.index == ['A', 'B', 'ITGB2', 'TNFRSF4']).all()
    np.testing.assert_almost_equal(mdata.mod['metabolite'].X.mean(), -0.18889697, decimal=5)
    np.testing.assert_almost_equal(mdata.mod['receptor'].X.mean(), 0.7754228, decimal=5)



from liana.method.sc._Method import Method, MethodMeta
import numpy as np

def _inter_score(x):
    inter_score = np.minimum(x['ligand_cdf'], x['receptor_cdf'])
    return inter_score, None

_scseqcomm = MethodMeta(method_name="scSeqComm",
                        complex_cols=["ligand_means", "receptor_means"],
                        add_cols=["ligand_cdf", "receptor_cdf"],
                        fun=_inter_score,
                        magnitude="inter_score",
                        magnitude_ascending=False,
                        specificity=None,
                        specificity_ascending=None,
                        permute=False,
                        reference="Baruzzo, G., Cesaro, G., Di Camillo, B. "
                                  "2022. Identify, quantify and characterize cellular communication "
                                  "from single-cell RNA-sequencing data with scSeqComm. Bioinformatics, "
                                  "38(7), pp.1920-1929"
                        )
scseqcomm = Method(_method=_scseqcomm)



from liana.resource import translate_column, translate_resource, get_hcop_orthologs, select_resource

import pandas as pd

def test_complex_cases():
    map_df = pd.DataFrame(
        {"source":
            ["CSF2RA", "IFNL3", "IFNL3", "IFNLR1", "IL10RB", "HCST", "CD8A", "CD8B", "IL4"],
        "target":
            ["Csf2ra", "Ifnl3", "Ifnl2", "Ifnlr1", "Il10rb", "Hcst", "Cd8a", "Cd8b1", "Il4"]
            }
    )
    df = pd.DataFrame(
        {
            "symbol": [
                "CSF2RA_CSF2RB",  # one to many
                "IFNL3_IFNLR1_IL10RB",  # 3 subunits
                "HCST_KLRK1",  # one subunit missing
                "CD8A_CD8B",  # 1 to 1
                "IL4",  # 1 to 1 simple protein
            ]
        }
    )

    default = translate_column(
        df,
        map_df=map_df,
        column="symbol",
    )
    assert all(default["symbol"] == ["Cd8a_Cd8b1", "Il4"])

    to_many = translate_column(
        df,
        map_df=map_df,
        column="symbol",
        replace=True,
        one_to_many=2,
    )
    expected = [
        "Cd8a_Cd8b1",
        "Ifnl2_Ifnlr1_Il10rb",
        "Ifnl3_Ifnlr1_Il10rb",
        "Il4",
    ]

    assert to_many.shape == (4, 1)
    assert all(to_many["symbol"].isin(expected))

    keep_missing = translate_column(
        df,
        map_df=map_df,
        column="symbol",
        replace=False,
        one_to_many=2,
    )
    untranslated = keep_missing["symbol"].isin(["HCST_KLRK1"])
    assert untranslated.any()


def test_translate_resource():
    resource = select_resource()
    map_df = get_hcop_orthologs(columns=['human_symbol', 'mouse_symbol'], min_evidence=3)
    map_df = map_df.rename(columns={"human_symbol": "source", "mouse_symbol": "target"})

    translated = translate_resource(resource, map_df, one_to_many=1)
    assert translated.shape[0] > 3000
    translated2 = translate_resource(resource, map_df, one_to_many=5, replace=False)
    assert translated2.shape[0] > translated.shape[0]


def test_get_hcop():
    mapping = get_hcop_orthologs(columns=None, min_evidence=0)
    assert mapping.shape[0] > 1000
    assert mapping.shape[1] == 16 # 15 columns + added evidence column



import pandas
from numpy import max, min
from numpy.testing import assert_almost_equal
from pandas import DataFrame

from liana.method import cellphonedb, singlecellsignalr as sca, \
    natmi, connectome, logfc, geometric_mean, cellchat, scseqcomm

from liana.testing._sample_anndata import generate_toy_adata

# load toy adata
adata = generate_toy_adata()
expected_shape = adata.shape


def test_cellchat():
    cellchat(adata, groupby='bulk_labels', use_raw=True, n_perms=4)

    assert adata.shape == expected_shape
    assert 'liana_res' in adata.uns.keys()

    liana_res = adata.uns['liana_res']
    assert isinstance(liana_res, pandas.DataFrame)

    assert 'lr_probs' in liana_res.columns
    assert 'cellchat_pvals' in liana_res.columns
    assert max(liana_res[(liana_res.ligand == "TIMP1")].lr_probs) == 0.20561589810421071
    assert liana_res[liana_res['receptor_complex']=='CD74_CXCR4']['lr_probs'].max() == 0.10198416583005679
    assert liana_res[liana_res['receptor_complex']=='CD74_CXCR4']['cellchat_pvals'].mean() == 0.5125


def test_cellphonedb():
    cellphonedb(adata, groupby='bulk_labels', use_raw=True, n_perms=4)

    assert adata.shape == expected_shape
    assert 'liana_res' in adata.uns.keys()

    liana_res = adata.uns['liana_res']
    assert isinstance(liana_res, pandas.DataFrame)

    assert 'lr_means' in liana_res.columns
    assert 'cellphone_pvals' in liana_res.columns
    assert_almost_equal(max(liana_res[(liana_res.ligand == "TIMP1")].lr_means), 2.134743630886078, decimal=6)
    assert_almost_equal(liana_res[liana_res['receptor_complex']=='CD74_CXCR4']['lr_means'].max(), 1.4035000205039978, decimal=6)
    assert liana_res[liana_res['receptor_complex']=='CD74_CXCR4']['cellphone_pvals'].mean() == 0.415


def test_cellphonedb_none():
    cellphonedb(adata, groupby='bulk_labels', use_raw=True, n_perms=None)
    assert adata.shape == expected_shape
    liana_res = adata.uns['liana_res']
    assert_almost_equal(liana_res[liana_res['receptor_complex']=='CD74_CXCR4']['lr_means'].max(), 1.4035000205039978, decimal=6)
    assert 'cellphone_pvals' not in liana_res.columns


def test_geometric_mean():
    geometric_mean(adata, groupby='bulk_labels', use_raw=True, n_perms=4, n_jobs=2)

    assert adata.shape == expected_shape
    assert 'liana_res' in adata.uns.keys()

    liana_res = adata.uns['liana_res']
    assert isinstance(liana_res, pandas.DataFrame)

    assert 'lr_gmeans' in liana_res.columns
    assert 'gmean_pvals' in liana_res.columns
    assert_almost_equal(max(liana_res[(liana_res.ligand == "TIMP1")].lr_gmeans), 2.126363309240465, decimal=6)
    assert_almost_equal(liana_res[liana_res['receptor_complex']=='CD74_CXCR4']['lr_gmeans'].max(), 1.4016519940029961, decimal=6)
    assert liana_res[liana_res['receptor_complex']=='CD74_CXCR4']['gmean_pvals'].mean() == 0.5125


def test_natmi():
    natmi(adata, groupby='bulk_labels', use_raw=True)

    assert adata.shape == expected_shape
    assert 'liana_res' in adata.uns.keys()

    liana_res = adata.uns['liana_res']
    assert isinstance(liana_res, pandas.DataFrame)

    assert 'expr_prod' in liana_res.columns
    assert 'spec_weight' in liana_res.columns
    assert_almost_equal(max(liana_res[(liana_res.ligand == "TIMP1")].spec_weight), 0.0604750001773605, decimal=6)
    assert_almost_equal(max(liana_res[(liana_res.ligand == "TIMP1")].expr_prod), 4.521420922884062, decimal=6)
    assert_almost_equal(liana_res[liana_res['receptor_complex']=='CD74_CXCR4']['spec_weight'].max(), 0.03480120361979308, decimal=6)
    assert_almost_equal(liana_res[liana_res['receptor_complex']=='CD74_CXCR4']['expr_prod'].max(), 1.9646283122925752, decimal=6)

def test_scseqcomm():
    scseqcomm(adata, groupby='bulk_labels', use_raw=True, expr_prop = 0, return_all_lrs=True)

    assert adata.shape == expected_shape
    assert 'liana_res' in adata.uns.keys()

    liana_res = adata.uns['liana_res']
    assert isinstance(liana_res, pandas.DataFrame)

    assert 'inter_score' in liana_res.columns

    assert_almost_equal(liana_res[(liana_res.ligand == "TIMP1") & \
                                (liana_res.receptor == "CD63") & \
                                (liana_res.source == "Dendritic") & \
                                (liana_res.target == "CD4+/CD45RA+/CD25- Naive T")]['inter_score'].values, 0.6819619345, decimal = 5)
    assert_almost_equal(max(liana_res[(liana_res.receptor_complex == "CD74_CXCR4")]['inter_score']), 1, decimal = 6)


def test_sca():
    sca(adata, groupby='bulk_labels', use_raw=True)

    assert adata.shape == expected_shape
    assert 'liana_res' in adata.uns.keys()

    liana_res = adata.uns['liana_res']
    assert isinstance(liana_res, pandas.DataFrame)

    assert 'lrscore' in liana_res.columns
    assert_almost_equal(max(liana_res[(liana_res.ligand == "TIMP1")].lrscore), 0.781133536891427, decimal=6)
    assert_almost_equal(liana_res[liana_res['receptor_complex']=='CD74_CXCR4']['lrscore'].max(), 0.7017243729003677, decimal=6)


def test_logfc():
    logfc(adata, groupby='bulk_labels', use_raw=True)

    assert adata.shape == expected_shape
    assert 'liana_res' in adata.uns.keys()

    liana_res = adata.uns['liana_res']
    assert isinstance(liana_res, pandas.DataFrame)

    assert 'lr_logfc' in liana_res.columns
    assert_almost_equal(max(liana_res[(liana_res.ligand == "TIMP1")].lr_logfc), 1.4352725744247437, decimal=6)
    assert_almost_equal(liana_res[liana_res['receptor_complex']=='CD74_CXCR4']['lr_logfc'].max(), 1.0422011613845825, decimal=6)


def test_connectome():
    connectome(adata, groupby='bulk_labels', use_raw=True)
    assert adata.shape == expected_shape
    assert 'liana_res' in adata.uns.keys()

    liana_res = adata.uns['liana_res']
    assert isinstance(liana_res, pandas.DataFrame)

    assert 'expr_prod' in liana_res.columns
    assert 'scaled_weight' in liana_res.columns
    assert_almost_equal(max(liana_res[(liana_res.ligand == "TIMP1")].scaled_weight), 0.9669451713562012, decimal=6)
    assert_almost_equal(max(liana_res[(liana_res.ligand == "TIMP1")].expr_prod), 4.521420922884062, decimal=6)
    assert_almost_equal(liana_res[liana_res['receptor_complex']=='CD74_CXCR4']['scaled_weight'].max(), 0.9002860486507416, decimal=6)
    assert_almost_equal(liana_res[liana_res['receptor_complex']=='CD74_CXCR4']['expr_prod'].max(), 1.9646283122925752, decimal=6)



def test_with_all_lrs():
    natmi(adata, groupby='bulk_labels', use_raw=True, return_all_lrs=True, key_added='all_res')
    lr_all = adata.uns['all_res']
    assert lr_all.shape == (4200, 15)
    assert all(lr_all[~lr_all.lrs_to_keep][natmi.magnitude] == min(lr_all[natmi.magnitude])) is True
    assert all(lr_all[~lr_all.lrs_to_keep][natmi.specificity] == min(lr_all[natmi.specificity])) is True


def test_methods_by_sample():
    logfc.by_sample(adata, groupby='bulk_labels', use_raw=True, return_all_lrs=True, sample_key='sample')
    lr_by_sample = adata.uns['liana_res']

    assert 'sample' in lr_by_sample.columns
    assert lr_by_sample.shape == (10836, 15)


def test_methods_on_mdata():
    from liana.testing._sample_anndata import generate_toy_mdata
    from itertools import product

    mdata = generate_toy_mdata()
    mdata.mod['adata_y'].var.index = 'scaled:' + mdata.mod['adata_y'].var.index
    interactions = list(product(mdata.mod['adata_x'].var.index, mdata.mod['adata_y'].var.index))
    interactions = interactions[0:10]

    sca(mdata,
        groupby='bulk_labels',
        n_perms=None,
        use_raw=False,
        interactions=interactions,
        verbose=True,
        mdata_kwargs=dict(
            x_mod='adata_x',
            y_mod='adata_y',
            x_transform=False,
            y_transform=False
            ),
        )

    assert mdata.uns['liana_res'].shape == (132, 12)

def test_wrong_resource():
    from pytest import raises
    with raises(ValueError):
        natmi(adata, resource_name='mouseconsensus', groupby='bulk_labels', use_raw=True, n_perms=4)

    with raises(ValueError):
        natmi(adata, interactions=[('x', 'D')], groupby='bulk_labels', use_raw=True, n_perms=4)

    with raises(ValueError):
        resource = DataFrame({'ligand': ['A', 'B'], 'receptor': ['C', 'D']})
        natmi(adata, resource=resource, groupby='bulk_labels', use_raw=True, n_perms=4)



from __future__ import annotations
from typing import Union, Optional

import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix, isspmatrix_csr
from anndata import AnnData
from mudata import MuData

from liana.method._pipe_utils._common import _get_props
from liana.utils.mdata_to_anndata import mdata_to_anndata
from liana.resource.select_resource import _handle_resource
from liana.method._pipe_utils import prep_check_adata, assert_covered

from liana.method.sp._utils import _add_complexes_to_var, _zscore
from liana.method.sp._bivariate._global_functions import GlobalFunction
from liana.method.sp._bivariate._local_functions import LocalFunction

from liana._logging import _logg
from liana._docs import d
from liana._constants import Keys as K, DefaultValues as V


class SpatialBivariate():
    """ A class for bivariate local spatial metrics. """
    def __init__(self, x_name: str = 'x', y_name: str = 'y'):
        self.x_name = x_name
        self.y_name = y_name

    @d.dedent
    def __call__(self,
                 mdata: (MuData | AnnData),
                 local_name: (str | None) = 'cosine',
                 global_name: (None | str | list) = None,
                 resource_name: str = None,
                 resource: Optional[pd.DataFrame] = V.resource,
                 interactions: list = V.interactions,
                 connectivity_key: str = K.connectivity_key,
                 mask_negatives: bool = False,
                 add_categories: bool = False,
                 n_perms: int = None,
                 seed: int = V.seed,
                 nz_prop: float = 0.05,
                 remove_self_interactions: bool = True,
                 complex_sep: (None | str) = "_",
                 xy_sep: str = V.lr_sep,
                 inplace: bool = V.inplace,
                 key_added: str = "local_scores",
                 verbose: bool = V.verbose,
                 **kwargs
                 ) -> Union[AnnData, pd.DataFrame] | None:
        """
        A method for bivariate local spatial metrics.

        Parameters
        ----------

        %(mdata)s
        %(local_name)s
        %(global_name)s
        %(interactions)s
        %(resource)s
        %(resource_name)s
        %(connectivity_key)s
        %(mask_negatives)s
        %(add_categories)s
        %(n_perms)s
        %(seed)s
        nz_prop: float
            Minimum proportion of non-zero values for each features. For example, if working with gene expression data,
            this would be the proportion of cells expressing a gene. Both features must have a proportion greater than
            `nz_prop` to be considered in the analysis.
        complex_sep: str
            Separator to use for complex names.
        xy_sep: str
            Separator to use for interaction names.
        remove_self_interactions: bool
            Whether to remove self-interactions. `True` by default.
        %(inplace)s
        key_added: str
            Key in `mdata.mod` (if MuData) or `adata.obsm` (if AnnData) where the local scores will be stored.
        %(verbose)s

        **kwargs : dict, optional
            Additional keyword arguments:
            - For AnnData:
                %(x_name)s By default: 'ligand'.
                %(y_name)s By default: 'receptor'.

            - For MuData:
                %(x_mod)s
                %(y_mod)s
                %(x_name)s By default: 'x'.
                %(y_name)s By default: 'y'.
                x_use_raw: bool
                    Whether to use the raw counts for the x-mod.
                y_use_raw: bool
                    Whether to use the raw counts for y-mod.
                x_layer: str
                    Layer to use for x-mod.
                y_layer: str
                    Layer to use for y-mod.
                x_transform: bool
                    Function to transform the x-mod.
                y_transform: bool
                    Function to transform the y-mod.

        Returns
        -------
        If `inplace` is `True`, the results are added to `mdata` and `None` is returned.
        Note that `var`, `obs`, `obsm`, `uns` and `obsp` attributes are copied to the output object.
        When `mdata` is an `AnnData` object, the results are stored in `mdata.obsm[key_added]`
        When `mdata` is an `MuData` object, the results are stored in `mdata.mod[key_added]`

        `AnnData` objects in `obsm` will not be copied to the output object.

        If `inplace` is `False`, the results are returned.

        """

        if n_perms is not None:
            if not isinstance(n_perms, int) or n_perms < 0:
                raise ValueError("n_perms must be None, 0 for analytical or > 0 for permutation")
        if global_name is not None:
            if isinstance(global_name, str):
                global_name = [global_name]
        if (n_perms == 0) and ((local_name not in ["morans", None]) or ~np.isin(global_name, ["morans", None]).any()):
            raise ValueError("An analytical solution is currently available only for Moran's R")

        if local_name is not None:
            local_fun = LocalFunction._get_instance(name=local_name)

        if isinstance(mdata, MuData):
            adata = self._process_mudata(mdata, complex_sep, verbose=verbose, **kwargs)
        elif isinstance(mdata, AnnData):
            adata = self._process_anndata(mdata, complex_sep, verbose=verbose, **kwargs)
        else:
            raise ValueError("Invalid type, `adata/mdata` must be an AnnData/MuData object")

        resource = _handle_resource(interactions=interactions,
                                    resource=resource,
                                    resource_name=resource_name,
                                    x_name=self.x_name,
                                    y_name=self.y_name,
                                    verbose=verbose
                                    )
        weight = self._handle_connectivity(adata=adata, connectivity_key=connectivity_key)

        if complex_sep is not None:
            adata = _add_complexes_to_var(adata,
                                          np.union1d(resource[self.x_name].astype(str),
                                                     resource[self.y_name].astype(str)
                                                     ),
                                          complex_sep=complex_sep
                                          )

        # filter_resource
        resource = resource[(np.isin(resource[self.x_name], adata.var_names)) &
                            (np.isin(resource[self.y_name], adata.var_names))]

        self_interactions = resource[self.x_name] == resource[self.y_name]
        if self_interactions.any() & remove_self_interactions:
            _logg(f"Removing {self_interactions.sum()} self-interactions", verbose=verbose)
            resource = resource[~self_interactions]

        # get entities
        entities = np.union1d(np.unique(resource[self.x_name]),
                                np.unique(resource[self.y_name]))
        assert_covered(entities, adata.var_names, verbose=verbose)

        # Filter to only include the relevant features
        adata = adata[:, np.intersect1d(entities, adata.var.index)]

        xy_stats = pd.DataFrame({'means': adata.X.mean(axis=0).A.flatten(),
                                 'props': _get_props(adata.X)},
                                index=adata.var_names
                                ).reset_index().rename(columns={'index': 'gene'})
        # join global stats to LRs from resource
        xy_stats = resource.merge(self._rename_means(xy_stats, entity=self.x_name)).merge(
                                    self._rename_means(xy_stats, entity=self.y_name))

        # filter according to props
        xy_stats = xy_stats[(xy_stats[f'{self.x_name}_props'] >= nz_prop) &
                            (xy_stats[f'{self.y_name}_props'] >= nz_prop)]
        if xy_stats.empty:
            raise ValueError("No features with non-zero proportions")

        # create interaction column
        xy_stats['interaction'] = xy_stats[self.x_name] + xy_sep + xy_stats[self.y_name]

        x_mat = adata[:, xy_stats[self.x_name]].X
        y_mat = adata[:, xy_stats[self.y_name]].X

        if global_name is not None:
            for gname in global_name:
                global_fun = GlobalFunction.instances[gname]
                global_fun(xy_stats,
                           x_mat=x_mat,
                           y_mat=y_mat,
                           weight=weight,
                           seed=seed,
                           n_perms=n_perms,
                           mask_negatives=mask_negatives,
                           verbose=verbose,
                           )

        if local_name is None:
            return xy_stats

        # Calculate local scores
        if add_categories or mask_negatives:
            local_cats = self._categorize(x_mat=x_mat,
                                          y_mat=y_mat,
                                          weight=weight,
                                          )
        else:
            local_cats = None

        # get local scores
        local_scores, local_pvals = \
            local_fun(x_mat=x_mat,
                      y_mat=y_mat,
                      weight=weight,
                      seed=seed,
                      n_perms=n_perms,
                      mask_negatives=mask_negatives,
                      verbose=verbose,
                      )

        xy_stats.loc[:, ['mean', 'std']] = \
            np.vstack(
                [np.mean(local_scores, axis=0),
                 np.std(local_scores, axis=0)]
                ).T

        if mask_negatives:
            local_scores = np.where(local_cats!=1, 0, local_scores)
            if local_pvals is not None:
                local_pvals = np.where(local_cats!=1, 1, local_pvals)

        local_scores = AnnData(csr_matrix(local_scores),
                               obs=adata.obs,
                               var=xy_stats.set_index('interaction'),
                               uns=adata.uns,
                               obsm=adata.obsm,
                               obsp=adata.obsp,
                               )

        if add_categories:
            local_scores.layers['cats'] = csr_matrix(local_cats)
        if local_pvals is not None:
            local_scores.layers['pvals'] = csr_matrix(local_pvals)

        return self._handle_return(mdata, xy_stats, local_scores, key_added, inplace)

    def _process_anndata(self,
                         adata,
                         complex_sep,
                         verbose,
                         **kwargs):
        expected_params = {'x_name', 'y_name', 'use_raw', 'layer'}
        self.validate_kwargs(expected_params=expected_params, **kwargs)

        self.x_name = kwargs.get('x_name', 'ligand')
        self.y_name = kwargs.get('y_name', 'receptor')

        return prep_check_adata(adata=adata,
                                use_raw=kwargs.get('use_raw', V.use_raw),
                                layer=kwargs.get('layer', V.layer),
                                verbose=verbose,
                                obsm=adata.obsm.copy(),
                                uns=adata.uns.copy(),
                                groupby=None,
                                min_cells=None,
                                complex_sep=complex_sep,
                                )

    def _process_mudata(self,
                        mdata,
                        complex_sep,
                        verbose,
                        **kwargs):
        expected_params = {'x_name', 'y_name',
                           'x_mod', 'y_mod',
                           'x_use_raw', 'x_layer',
                           'y_use_raw', 'y_layer',
                           'x_transform', 'y_transform'}
        self.validate_kwargs(expected_params=expected_params, **kwargs)

        self.x_name = kwargs.get('x_name', self.x_name)
        self.y_name = kwargs.get('y_name', self.y_name)

        x_mod = kwargs.get('x_mod')
        y_mod = kwargs.get('y_mod')

        if x_mod is None or y_mod is None:
            raise ValueError("MuData processing requires 'x_mod' and 'y_mod' parameters.")

        adata = mdata_to_anndata(mdata,
                                 x_mod=x_mod,
                                 y_mod=y_mod,
                                 x_use_raw=kwargs.get('x_use_raw', V.use_raw),
                                 x_layer=kwargs.get('x_layer', V.layer),
                                 y_use_raw=kwargs.get('y_use_raw', V.use_raw),
                                 y_layer=kwargs.get('y_layer', V.layer),
                                 x_transform=kwargs.get('x_transform', False),
                                 y_transform=kwargs.get('y_transform', False),
                                 verbose=verbose
                                 )

        return prep_check_adata(adata=adata,
                                use_raw=False,
                                layer=None,
                                verbose=verbose,
                                obsm = adata.obsm.copy(),
                                uns=adata.uns.copy(),
                                groupby=None,
                                min_cells=None,
                                complex_sep=complex_sep, # NOTE
                                )


    def validate_kwargs(self, expected_params, **kwargs):
        unexpected_kwargs = set(kwargs) - expected_params
        if unexpected_kwargs:
            raise ValueError(f"Unexpected keyword arguments: {unexpected_kwargs}")


    def _rename_means(self, lr_stats, entity):
        df = lr_stats.copy()
        df.columns = df.columns.map(lambda x: entity + '_' + str(x) if x != 'gene' else 'gene')
        return df.rename(columns={'gene': entity})

    def _handle_return(self, data, stats, local_scores, key_added, inplace=False):
        if not inplace:
            return stats, local_scores

        if isinstance(data, MuData):
            data.mod[key_added] = local_scores
        else:
            data.obsm[key_added] = local_scores

    def _handle_connectivity(self, adata, connectivity_key):
        if connectivity_key not in adata.obsp.keys():
            raise ValueError(f'No connectivity matrix founds in mdata.obsp[{connectivity_key}]')
        connectivity = adata.obsp[connectivity_key]

        if not isspmatrix_csr(connectivity) or (connectivity.dtype != np.float32):
            connectivity = csr_matrix(connectivity, dtype=np.float32)

        return connectivity

    def _encode_cats(self, a, weight):
        if np.all(a >= 0):
            a = _zscore(a)
        a = weight @ a
        a = np.where(a > 0, 1, np.where(a < 0, -1, np.nan))
        return a

    def _categorize(self, x_mat, y_mat, weight):
        x_cats = self._encode_cats(x_mat.A, weight)
        y_cats = self._encode_cats(y_mat.A, weight)
        cats = x_cats + y_cats
        cats = np.where(cats == 2, 1, np.where(cats == 0, -1, 0))

        return cats

    def show_functions(self):
        """
        Print information about all bivariate local metrics.
        """
        funs = LocalFunction.instances.copy()
        for function in funs.values():
            funs[function.name] = {
                "metadata":function.metadata,
                "reference":function.reference,
                }
        return pd.DataFrame(funs).T.reset_index().rename(columns={"index":"name"})

bivariate = SpatialBivariate()



import pandas as pd
from liana.resource.get_metalinks import get_metalinks, get_metalinks_values, describe_metalinks
import pathlib

# TODO: replace with link, None should use the link, optionally use path
import os
filepath = pathlib.Path(__file__).parent.absolute()
os.chdir(filepath)

def test_get_metalinks():
    # set path to here
    result = get_metalinks(tissue_location='Brain',
                           hmdb_ids='HMDB0000073',
                           uniprot_ids='P14416'
                           ).drop_duplicates(['hmdb', 'uniprot'])
    assert isinstance(result, pd.DataFrame)
    assert result.shape == (1, 8)
    assert 'Dopamine' in result['metabolite'].values
    assert 'HMDB0000073' in result['hmdb'].values
    assert 'P14416' in result['uniprot'].values
    assert 'DRD2' in result['gene_symbol'].values


def test_get_metalinks_values():
    # Call the function with a test db_path, table_name, and column_name
    result = get_metalinks_values('disease', 'disease')

    # Check that the result is a list
    assert isinstance(result, list)

    assert 'Obesity' in result
    assert 'Schizophrenia' in result
    assert len(result) == 567


def test_describe_metalinks():
    out = describe_metalinks(return_output=True)
    assert 'metabolites' in out
    assert 'proteins' in out
    assert 'edges' in out
    assert 'Column ID: 8' in out



from __future__ import annotations

from liana.method.sc._Method import MethodMeta
from liana.method.sc._liana_pipe import liana_pipe
from liana._docs import d
from mudata import MuData
from liana._constants import Keys as K, DefaultValues as V

import anndata as an
from pandas import DataFrame
from typing import Optional


class AggregateClass(MethodMeta):
    """LIANA's Method Consensus Class"""
    def __init__(self, _SCORE, methods):
        super().__init__(method_name=_SCORE.method_name,
                         complex_cols=[],
                         add_cols=[],
                         fun=_SCORE.fun,
                         magnitude=_SCORE.magnitude,
                         magnitude_ascending=True,
                         specificity=_SCORE.specificity,
                         specificity_ascending=True,
                         permute=_SCORE.permute,
                         reference=_SCORE.reference
                         )
        self._SCORE = _SCORE
        self.methods = methods

        # Define sc to aggregate
        self.specificity_specs = {method.method_name: (
            method.specificity, method.specificity_ascending) for method in methods
            if method.specificity is not None}
        self.magnitude_specs = {method.method_name: (
            method.magnitude, method.magnitude_ascending) for method in methods
            if method.magnitude is not None}

        # Define additional columns needed depending on the methods to be run
        self.add_cols = list(
            {x for li in [method.add_cols for method in methods] for x in li}
        )
        self.complex_cols = list(
            {x for li in [method.complex_cols for method in methods] for x in li}
        )

    def describe(self):
        """Briefly described the method"""
        print(
            f"{self.method_name} returns `{self.magnitude}`, `{self.specificity}`. "
            f"{self.magnitude} and {self.specificity} respectively represent an aggregate of the "
            f"`magnitude`- and `specificity`-related scoring functions from the different methods."
        )

    @d.dedent
    def __call__(self,
                 adata: an.AnnData | MuData,
                 groupby: str,
                 resource_name: str = V.resource_name,
                 expr_prop: float = V.expr_prop,
                 min_cells: int = V.min_cells,
                 groupby_pairs: Optional[DataFrame] = V.groupby_pairs,
                 base: float = V.logbase,
                 aggregate_method: str = 'rra',
                 consensus_opts: Optional[list] = None,
                 return_all_lrs: bool = V.return_all_lrs,
                 key_added: str = K.uns_key,
                 use_raw: Optional[bool] = V.use_raw,
                 layer: Optional[str] = V.layer,
                 de_method: str = V.de_method,
                 n_perms: int = V.n_perms,
                 seed: int = V.seed,
                 n_jobs: int = 1,
                 resource: Optional[DataFrame] = V.resource,
                 interactions: Optional[list] = V.interactions,
                 mdata_kwargs: dict = dict(),
                 inplace: bool = V.inplace,
                 verbose: Optional[bool] = V.verbose,
                 ):
        """
        Get an aggregate of ligand-receptor scores from multiple methods.

        Parameters
        ----------
        %(adata)s
        %(groupby)s
        %(resource_name)s
        %(expr_prop)s
        %(min_cells)s
        %(groupby_pairs)s
        %(base)s
        aggregate_method
            Method aggregation approach, one of ['mean', 'rra'], where `mean` represents the
            mean rank, while 'rra' is the RobustRankAggregate (Kolde et al., 2014)
            of the interactions
        consensus_opts
            Strategies to aggregate interactions across methods.
            Default is None - i.e. ['Specificity', 'Magnitude'] and both specificity and magnitude are aggregated.
        %(return_all_lrs)s
        %(key_added)s
        %(use_raw)s
        %(layer)s
        %(de_method)s
        %(verbose)s
        %(n_perms_sc)s
        %(seed)s
        n_jobs
            Number of jobs to run in parallel.
        %(resource)s
        %(interactions)s
        %(mdata_kwargs)s
        %(inplace)s

        Returns
        -------
        If ``inplace = False``, returns a `DataFrame` with ligand-receptor results
        Otherwise, modifies the ``adata`` object with the following key:
            - :attr:`anndata.AnnData.uns` ``['liana_res']`` with the aforementioned DataFrame
        """

        liana_res = liana_pipe(adata=adata,
                               groupby=groupby,
                               resource_name=resource_name,
                               resource=resource,
                               groupby_pairs = groupby_pairs,
                               interactions=interactions,
                               expr_prop=expr_prop,
                               min_cells=min_cells,
                               base=base,
                               return_all_lrs=return_all_lrs,
                               de_method=de_method,
                               verbose=verbose,
                               _score=self,
                               use_raw=use_raw,
                               layer=layer,
                               n_perms=n_perms,
                               seed=seed,
                               n_jobs=n_jobs,
                               _methods=self.methods,
                               _aggregate_method=aggregate_method,
                               _consensus_opts=consensus_opts,
                               mdata_kwargs=mdata_kwargs
                               )

        if inplace:
            adata.uns[key_added] = liana_res
        return None if inplace else liana_res

_rank_aggregate_meta = \
    MethodMeta(method_name="Rank_Aggregate",
               complex_cols=[],
               add_cols=[],
               fun=None,  # change to _robust_rank
               magnitude='magnitude_rank',
               magnitude_ascending=True,
               specificity='specificity_rank',
               specificity_ascending=True,
               permute=False,
               reference='Dimitrov, D., Türei, D., Garrido-Rodriguez, M., Burmedi, P.L., '
                         'Nagai, J.S., Boys, C., Ramirez Flores, R.O., Kim, H., Szalai, B., '
                         'Costa, I.G. and Valdeolivas, A., 2022. Comparison of methods and '
                         'resources for cell-cell communication inference from single-cell '
                         'RNA-Seq data. Nature Communications, 13(1), pp.1-13. '
               )



import numpy as np
from itertools import product

from liana.testing._sample_anndata import generate_toy_mdata, generate_toy_spatial, generate_anndata
from liana.testing._sample_resource import sample_resource
from liana.method.sp._bivariate._spatial_bivariate import bivariate
from liana._constants import DefaultValues as V
from liana.utils.transform import zi_minmax

mdata = generate_toy_mdata()
interactions = list(product(mdata.mod['adata_x'].var.index,
                            mdata.mod['adata_y'].var.index))
ones = np.ones((mdata.shape[0], mdata.shape[0]), dtype=np.float64)
mdata.obsp['ones'] = ones


def test_bivar_morans_perms():
    bivariate(mdata,
              x_mod='adata_x',
              y_mod='adata_y',
              local_name='morans',
              n_perms=2,
              nz_prop=0,
              x_use_raw=False,
              y_use_raw=False,
              interactions=interactions)


    assert 'local_scores' in mdata.mod.keys()
    local_pvals = mdata.mod['local_scores'].layers['pvals']
    np.testing.assert_almost_equal(mdata.mod['local_scores'].X.sum(), -346.55872, decimal=4)
    np.testing.assert_almost_equal(np.mean(local_pvals), 0.52787581, decimal=6)


def test_bivar_nondefault():
    global_stats, local_scores = \
          bivariate(mdata,
                x_mod='adata_x',
                y_mod='adata_y',
                local_name='morans',
                global_name=['morans', 'lee'],
                n_perms=0,
                nz_prop=0,
                connectivity_key='ones',
                remove_self_interactions=False,
                x_layer = "scaled",
                y_layer = "scaled",
                x_use_raw=False,
                y_use_raw=False,
                inplace=False,
                add_categories=True,
                interactions=interactions
                )

    np.testing.assert_almost_equal(global_stats['morans'].sum(), 0)
    np.testing.assert_almost_equal(global_stats['lee'].sum(), 0)
    assert global_stats['lee_pvals'].unique()[0] is None
    assert 'morans_pvals' in global_stats.columns

    local_scores.shape == (700, 100)
    np.testing.assert_almost_equal(np.min(np.min(local_scores.layers['pvals'])), 0.5, decimal=2)


def test_masked_spearman():
    bivariate(mdata,
              x_mod='adata_x',
              y_mod='adata_y',
              x_use_raw=False,
              y_use_raw=False,
              nz_prop=0,
              local_name='masked_spearman',
              interactions=interactions,
              connectivity_key='ones'
              )

    assert 'local_scores' in mdata.mod.keys()
    np.testing.assert_almost_equal(mdata.mod['local_scores'].X.mean(), 0.18438724, decimal=5)

    assert mdata.mod['local_scores'].var.shape == (90, 8)
    global_res = mdata.mod['local_scores'].var
    assert set(['mean','std']).issubset(global_res.columns)
    np.testing.assert_almost_equal(global_res['mean'].mean(), 0.18438746, decimal=5)
    np.testing.assert_almost_equal(global_res['std'].mean(), 8.498836e-07, decimal=5)


def test_vectorized_spearman():
    bivariate(mdata,
              x_mod='adata_x',
              y_mod='adata_y',
              x_use_raw=False,
              y_use_raw=False,
              local_name='spearman',
              nz_prop=0,
              n_perms=2,
              interactions=interactions,
              )
    assert 'local_scores' in mdata.mod.keys()
    bdata = mdata.mod['local_scores']
    np.testing.assert_almost_equal(bdata.X.mean(), 0.0077174882, decimal=5)
    np.testing.assert_almost_equal(bdata.layers['pvals'].mean(), 0.6153921568, decimal=3)

    assert mdata.mod['local_scores'].var.shape == (90, 8)
    global_res = mdata.mod['local_scores'].var
    assert set(['mean','std']).issubset(global_res.columns)
    np.testing.assert_almost_equal(global_res['mean'].mean(), 0.0077174, decimal=5)
    np.testing.assert_almost_equal(global_res['std'].mean(), 0.46906388, decimal=5)

### Test on AnnData and LRs
 # NOTE: these should be the same regardless of the local function
adata = generate_toy_spatial()
expected_gmorans = 0.0994394
expected_glee = 0.04854206

def test_morans_analytical():
    bivariate(adata,
              local_name='morans',
              global_name=['morans'],
              resource_name=V.resource_name,
              n_perms=0,
              use_raw=True,
              mask_negatives=True
              )
    assert 'local_scores' in adata.obsm_keys()
    assert 'spatial' in adata.obsm.keys()
    assert 'louvain' in adata.uns.keys()
    assert 'spatial_connectivities' in adata.obsp.keys()

    lrdata = adata.obsm['local_scores']

    assert 'pvals' in lrdata.layers.keys()

    np.testing.assert_almost_equal(np.mean(lrdata[:,'MIF^CD74_CXCR4'].X), 0.12803833, decimal=6)
    np.testing.assert_almost_equal(np.mean(lrdata[:,'MIF^CD74_CXCR4'].layers['pvals']), 0.8764923, decimal=6)

    interaction = lrdata.var[lrdata.var.index == 'S100A9^ITGB2']
    np.testing.assert_almost_equal(interaction['morans'].values, expected_gmorans)
    np.testing.assert_almost_equal(interaction['morans_pvals'].values, 3.4125671e-07)

def test_cosine_permutation():
    adata.layers['array'] = adata.raw.X.A
    bivariate(adata,
              local_name='cosine',
              global_name=['morans', 'lee'],
              resource_name='consensus',
              n_perms=100,
              use_raw=False,
              layer='array'
              )
    lrdata = adata.obsm['local_scores']

    assert 'pvals' in lrdata.layers.keys()

    np.testing.assert_almost_equal(lrdata[:,'MIF^CD74_CXCR4'].X.mean(), 0.32514292, decimal=6)
    np.testing.assert_almost_equal(np.mean(lrdata[:,'MIF^CD74_CXCR4'].layers['pvals']), 0.601228, decimal=4)

    interaction = lrdata.var[lrdata.var.index == 'S100A9^ITGB2']
    np.testing.assert_almost_equal(interaction['mean'].values, 0.56016606)
    np.testing.assert_almost_equal(interaction['std'].values, 0.33243373)
    np.testing.assert_almost_equal(interaction['morans'].values, expected_gmorans)
    np.testing.assert_almost_equal(interaction['morans_pvals'].values, 0.85)
    np.testing.assert_almost_equal(interaction['lee'].values, expected_glee)
    np.testing.assert_almost_equal(interaction['lee_pvals'].values, 0.93)


def test_jaccard_pval_none_cats():
    bivariate(adata,
              local_name='jaccard',
              global_name='lee',
              resource_name='consensus',
              n_perms=None,
              use_raw=True,
              add_categories=True
              )

    assert 'local_scores' in adata.obsm_keys()
    assert adata.obsm['local_scores'].var.shape == (32, 10)
    lrdata = adata.obsm['local_scores']

    assert 'cats' in lrdata.layers.keys()
    assert lrdata.layers['cats'].sum() == -6197
    assert 'pvals' not in adata.layers.keys()
    interaction = lrdata.var[lrdata.var.index == 'S100A9^ITGB2']
    np.testing.assert_almost_equal(interaction['lee'].values, expected_glee)

    np.testing.assert_almost_equal(lrdata[:,'S100A9^ITGB2'].X.mean(), 0.4117572, decimal=6)


def test_bivar_product():
    conn = mdata.obsp['spatial_connectivities']
    mdata.obsp['norm'] = conn / conn.sum(axis=1)
    bivariate(mdata,
              x_mod='adata_x',
              y_mod='adata_y',
              x_transform=zi_minmax,
              y_transform=zi_minmax,
              x_use_raw=False,
              y_use_raw=False,
              connectivity_key='norm',
              local_name='product',
              global_name=None,
              interactions=interactions,
              n_perms=None,
              add_categories=True,
              key_added='product'
              )
    assert 'product' in mdata.mod.keys()
    bdata = mdata.mod['product']
    assert 'cats' in bdata.layers.keys()
    assert bdata.uns is not None
    assert bdata.obsm is not None
    assert bdata.obsp is not None
    np.testing.assert_almost_equal(bdata.X.max(), 0.63145)
    assert 'lee' not in bdata.var.columns

def test_large_adata():
    adata = generate_anndata(n_obs=10001)
    resource = sample_resource(adata, n_lrs=20)
    bivariate(adata,
              resource=resource,
              local_name='pearson',
              global_name='morans',
              n_perms=None,
              use_raw=False,
              add_categories=False
              )
    lrdata = adata.obsm['local_scores']
    np.testing.assert_almost_equal(lrdata.X.mean(), 0.00048977, decimal=4)
    np.testing.assert_almost_equal(lrdata.var['morans'].mean(), 0.00012773558, decimal=4)


def test_wrong_interactions():
    from pytest import raises
    with raises(ValueError):
        bivariate(adata,
                 resource_name='mouseconsensus',
                 local_name='morans',
                 n_perms=None,
                 use_raw=True,
                 add_categories=True
                 )

def test_wrong_kwargs():
    from pytest import raises
    with raises(ValueError):
        bivariate(adata,
                 resource_name='mouseconsensus',
                 local_name='morans',
                 n_perms=None,
                 use_raw=True,
                 add_categories=True,
                 life='is good'
                 )


def test_show_bivariate():
    local_scores = bivariate.show_functions()
    assert local_scores.shape == (8, 3)



from liana.testing import _sample_dea
from liana.multi import df_to_lr
from liana.testing._sample_anndata import generate_toy_adata


# Create a toy AnnData object
adata = generate_toy_adata()
groupby = 'bulk_labels'
dea_df = _sample_dea(adata, groupby)


def test_dea_to_lr():
    lr_res = df_to_lr(adata,
                       dea_df=dea_df,
                       resource_name='consensus',
                       expr_prop=0.1,
                       groupby='bulk_labels',
                       stat_keys=['stat', 'pval', 'padjusted'],
                       use_raw=False,
                       complex_col=None,
                       verbose=True,
                       min_cells=10,
                       return_all_lrs=False,
                       )
    assert lr_res.shape == (374, 22)
    # assert ligand_stat, ligand_pval, and ligand_padjusted are in lr_res.columns
    columns = lr_res.columns
    expected_columns = ['ligand', 'ligand_stat', 'ligand_pval', 'ligand_padjusted', 'ligand_expr',
                        'receptor', 'receptor_stat', 'receptor_pval', 'receptor_padjusted', 'receptor_expr']
    for col in expected_columns:
        assert col in columns
    assert lr_res['interaction_padjusted'].mean() == 0.5540001846991026

def test_dea_to_lr_params():
    lr_res = df_to_lr(adata,
                      dea_df=dea_df,
                      expr_prop=0.1,
                      min_cells=10,
                      groupby='bulk_labels',
                      stat_keys=['stat', 'pval', 'padjusted'],
                      use_raw=False,
                      complex_col='stat',
                      verbose=True,
                      return_all_lrs=True,
                      )
    assert lr_res.shape == (3321, 23)
